"""State module for managing Disks."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    location: str,
    disk_name: str = None,
    resource_group_name: str = None,
    creation_data: make_dataclass(
        "CreationData",
        [
            ("create_option", str, field(default=None)),
            (
                "gallery_image_reference",
                make_dataclass(
                    "ImageDiskReference",
                    [
                        ("id", str, field(default=None)),
                        ("lun", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "image_reference",
                make_dataclass(
                    "ImageReference",
                    [
                        ("id", str, field(default=None)),
                        ("lun", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("logical_sector_size", int, field(default=None)),
            ("security_data_uri", str, field(default=None)),
            ("source_resource_id", str, field(default=None)),
            ("source_unique_id", str, field(default=None)),
            ("source_uri", str, field(default=None)),
            ("storage_account_id", str, field(default=None)),
            ("upload_size_bytes", int, field(default=None)),
        ],
    ) = None,
    extended_location: make_dataclass(
        "ExtendedLocation",
        [("name", str, field(default=None)), ("type", str, field(default=None))],
    ) = None,
    bursting_enabled: bool = None,
    data_access_auth_mode: make_dataclass(
        "DataAccessAuthMode",
        [
            ("azure_active_directory", str, field(default=None)),
        ],
    ) = None,
    disk_access_id: str = None,
    disk_iops_read_only: int = None,
    disk_iops_read_write: int = None,
    disk_mbps_read_only: int = None,
    disk_mbps_read_write: int = None,
    disk_size_gb: int = None,
    encryption: make_dataclass(
        "Encryption",
        [
            ("disk_encryption_set_id", str, field(default=None)),
            ("type", str, field(default=None)),
        ],
    ) = None,
    encryption_settings_collection: make_dataclass(
        "EncryptionSettingsCollection",
        [
            ("enabled", bool, field(default=None)),
            (
                "encryption_settings",
                make_dataclass(
                    "EncryptionSettings",
                    [
                        (
                            "disk_encryption_key",
                            make_dataclass(
                                "DiskEncryptionKey",
                                [
                                    ("secret_url", str, field(default=None)),
                                    (
                                        "source_vault",
                                        make_dataclass(
                                            "SourceVault",
                                            [("vm_uri", str, field(default=None))],
                                        ),
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "key_encryption_key",
                            make_dataclass(
                                "KeyEncryptionKey",
                                [
                                    ("key_url", str, field(default=None)),
                                    (
                                        "source_vault",
                                        make_dataclass(
                                            "SourceVault",
                                            [("vm_uri", str, field(default=None))],
                                        ),
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            ("encryption_settings_version", str, field(default=None)),
        ],
    ) = None,
    hyper_v_generation: str = None,
    max_shares: int = None,
    network_access_policy: str = None,
    os_type: str = None,
    public_network_access: str = None,
    security_profile: make_dataclass(
        "SecurityProfile",
        [
            ("secure_vm_disk_encryption_set_id", str, field(default=None)),
            ("security_type", str, field(default=None)),
        ],
    ) = None,
    tier: str = None,
    sku: make_dataclass(
        "DiskSku",
        [("name", str, field(default=None)), ("tier", str, field(default=None))],
    ) = None,
    tags: Dict[str, str] = None,
    zones: List[str] = None,
    resource_id: str = None,
    subscription_id: str = None,
) -> Dict:
    r"""Create or update Disks.

    Args:
        name(str): The identifier for this state.
        disk_name(str, Optional): The name of the managed disk that is being created.
        resource_group_name(str, Optional): The name of the resource group.
        creation_data(Dict[str, Any], Optional):
            Disk source information. CreationData information cannot be changed after the disk has been created.

            * create_option (str, Optional):
                Enumerates the possible sources of a disk's creation.
                    Enum type. Allowed values:
                        - "Attach" - Disk will be attached to a VM.
                        - "Copy" - Create a new disk or snapshot by copying from a disk or snapshot specified by the given sourceResourceId.
                        - "CopyStart" - Create a new disk by using a deep copy process, where the resource creation is considered complete only after all data has been copied from the source.
                        - "Empty" - Create an empty data disk of a size given by disk_size_gb.
                        - "FromImage" - Create a new disk from a platform image specified by the given imageReference or galleryImageReference.
                        - "Import" - Create a disk by importing from a blob specified by a sourceUri in a storage account specified by storageAccountId.
                        - "ImportSecure" - Similar to Import create option. Create a new Trusted Launch VM or Confidential VM supported disk by importing additional blob for VM guest state specified by securityDataUri in storage account specified by storageAccountId.
                        - "Restore" - Create a new disk by copying from a backup recovery point.
                        - "Upload" - Create a new disk by obtaining a write token and using it to directly upload the contents of the disk.
                        - "UploadPreparedSecure" - Similar to Upload create option. Create a new Trusted Launch VM or Confidential VM supported disk and upload using write token in both disk and VM guest state

            * gallery_image_reference(List(Dict[str, Any])):
                Required if creating from a Gallery Image. The id of the ImageDiskReference will be the ARM id of the shared galley image version from which to create a disk.
                    * id(str, Optional):
                        A relative uri containing either a Platform Image Repository or user image reference.
                    * lun(str, Optional):
                        If the disk is created from an image's data disk, this is an index that indicates which of the data disks in the image to use. For OS disks, this field is null.
            * image_reference(List(Dict[str, Any])):
                Disk source information.
                    * id(str, Optional):
                        A relative uri containing either a Platform Image Repository or user image reference.
                    * lun(str, Optional):
                        If the disk is created from an image's data disk, this is an index that indicates which of the data disks in the image to use. For OS disks, this field is null.
            * logical_sector_size(str, Optional):
                Logical sector size in bytes for Ultra disks. Supported values are 512 ad 4096. 4096 is the default.
            * security_data_url(str, Optional):
                If createOption is ImportSecure, this is the URI of a blob to be imported into VM guest state.
            * source_resource_id(str, Optional):
                If createOption is Copy, this is the ARM id of the source snapshot or disk.
            * source_unique_id(str, Optional):
                If this field is set, this is the unique id identifying the source of this resource.
            * source_uri(str, Optional):
                If createOption is Import, this is the URI of a blob to be imported into a managed disk.
            * storage_account_id(str, Optional):
                Required if createOption is Import. The Azure Resource Manager identifier of the storage account containing the blob to import as a disk.
            * upload_size_bytes(str, Optional):
                If createOption is Upload, this is the size of the contents of the upload including the VHD footer. This value should be between 20972032 (20 MiB + 512 bytes for the VHD footer) and 35183298347520 bytes (32 TiB + 512 bytes for the VHD footer).
        location(str, Optional): Resource location.
        os_type(str, Optional): the Operating System type.
        disk_size_gb(int, Optional): If creationData.createOption is Empty, this field is mandatory, and it indicates the size of the disk to create. If this field is present for updates or creation with other options, it indicates a resize. Resizes are only allowed if the disk is not attached to a running VM, and can only increase the disk's size.
        encryption_settings_collection(Dict[str, Any], Optional): encryptionSettingsCollection

            * enabled(bool, Optional): Set this flag to true and provide DiskEncryptionKey and optional KeyEncryptionKey to enable encryption. Set this flag to false and remove DiskEncryptionKey and KeyEncryptionKey to disable encryption. If EncryptionSettings is null in the request object, the existing settings remain unchanged.
            * encryption_settings(List[Dict[str, Any]], Optional): A collection of encryption settings, one for each disk volume.
                * disk_encryption_key(Dict[str, Any], Optional): diskEncryptionKey
                    * source_vault(Dict[str, Any], Optional): sourceVault
                        * id(str, Optional): Resource Id
                    * secret_url(str, Optional): Url pointing to a key or secret in KeyVault
                * key_encryption_key(Dict[str, Any], Optional): keyEncryptionKey
                    * source_vault(Dict[str, Any], Optional): sourceVault
                        * id(str, Optional): Resource Id
                    * key_url(str, Optional): Url pointing to a key or secret in KeyVault
            * encryption_settings_version(str, Optional): Describes what type of encryption is used for the disks. Once this field is set, it cannot be overwritten. '1.0' corresponds to Azure Disk Encryption with AAD app.'1.1' corresponds to Azure Disk Encryption.
        disk_iops_read_write(int, Optional): The number of IOPS allowed for this disk; only settable for UltraSSD disks. One operation can transfer between 4k and 256k bytes.
        disk_mbps_read_write(int, Optional): The bandwidth allowed for this disk; only settable for UltraSSD disks. MBps means millions of bytes per second - MB here uses the ISO notation, of powers of 10.
        disk_iops_read_only(int, Optional): The total number of IOPS that will be allowed across all VMs mounting the shared disk as ReadOnly. One operation can transfer between 4k and 256k bytes.
        disk_mbps_read_only(int, Optional): The total throughput (MBps) that will be allowed across all VMs mounting the shared disk as ReadOnly. MBps means millions of bytes per second - MB here uses the ISO notation, of powers of 10.
        hyper_v_generation(str, Optional): The hypervisor generation of the Virtual Machine. Applicable to OS disks only.
        max_shares(int, Optional): The maximum number of VMs that can attach to the disk at the same time. Value greater than one indicates a disk that can be mounted on multiple VMs at the same time.
        encryption(Dict[str, Any], Optional): encryption
            * disk_encryption_set_id(str, Optional): ResourceId of the disk encryption set to use for enabling encryption at rest.
            * type(str, Optional): The type of key used to encrypt the data of the disk.
        network_access_policy (str, Optional): Policy for accessing the disk via network.
        disk_access_id (str, Optional): ARM id of the DiskAccess resource for using private endpoints on disks.
        tier (str, Optional): Performance tier of the disk (e.g, P4, S10) as described here: https://azure.microsoft.com/en-us/pricing/details/managed-disks/. Does not apply to Ultra disks.
        extended_location(dict[str, Any], Optional): The extended location where the disk will be created. Extended location cannot be changed.
        bursting_enabled(bool, Optional): Set to true to enable bursting beyond the provisioned performance target of the disk. Bursting is disabled by default. Does not apply to Ultra disks.
        security_profile(): Contains the security related information for the resource.
        public_network_access(str, Optional): Policy for controlling export on the disk.
        data_access_auth_mode(str, Optional): Additional authentication requirements when exporting or uploading to a disk or snapshot.
        tags(Dict[str, str], Optional): Resource tags. Defaults to None.
        sku(Dict[str, Any], Optional): The disks sku name. Defaults to None.
            * name(str, Optional): The sku name.
            * tier(str, Optional): The sku tier.
        zones(List(str), Optional): The Logical zone list for Disk.
        resource_id(str, Optional): Disk resource id on Azure.
        subscription_id(str, Optional): Subscription credentials which uniquely identify Microsoft Azure subscription. The subscription ID forms part of the URI for every service call.


    Returns:
        Dict

    Examples:
        .. code-block:: sls

            my-disk:
              azure.compute.disks.present:
                - resource_group_name: resource-group-name
                - disk_name: disk-1
                - location: westus
                - creation_data:
                    create_option: FromImage
                    image_reference:
                      id: /Subscriptions/sub-id/Providers/Microsoft.Compute/Locations/westus/Publishers/Canonical/ArtifactTypes/VMImage/Offers/UbuntuServer/Skus/16.04-LTS/Versions/16.04.202109280
    """
    path_properties = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_group_name,
        "disk_name": disk_name,
    }
    query_properties = {}
    body_properties = {
        "location": location,
        "tags": tags,
        "sku": sku,
        "zones": zones,
        "extended_location": extended_location,
        "os_type": os_type,
        "hyper_v_generation": hyper_v_generation,
        "creation_data": creation_data,
        "disk_size_gb": disk_size_gb,
        "encryption_settings_collection": encryption_settings_collection,
        "disk_iops_read_write": disk_iops_read_write,
        "disk_mbps_read_write": disk_mbps_read_write,
        "disk_iops_read_only": disk_iops_read_only,
        "disk_mbps_read_only": disk_mbps_read_only,
        "encryption": encryption,
        "max_shares": max_shares,
        "network_access_policy": network_access_policy,
        "disk_access_id": disk_access_id,
        "tier": tier,
        "bursting_enabled": bursting_enabled,
        "security_profile": security_profile,
        "public_network_access": public_network_access,
        "data_access_auth_mode": data_access_auth_mode,
    }
    resource_idem_path = "azure.compute.disks"
    resource_id_format = "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/disks/{disk_name}"
    return await hub.tool.azure.generic.run_present(
        ctx,
        resource_idem_path,
        hub.tool.azure.api_versions.compute.disks,
        name,
        resource_id,
        resource_id_format,
        path_properties,
        query_properties,
        body_properties,
    )


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    disk_name: str = None,
    subscription_id: str = None,
) -> Dict:
    r"""Delete a Disk.

    Args:
        name(str): The identifier for this state.
        disk_name(str): The name of the disk.
        resource_group_name(str): The name of the resource group.
        subscription_id(str): Subscription Unique id.
        resource_id(str, Optional): Disk resource id on Azure.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.compute.disks.absent:
                - name: my-disk
                - resource_group_name: my-resource-group
                - disk_name: my-disk
                - subscription_id: my-subscription
    """
    path_properties = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_group_name,
        "disk_name": disk_name,
    }
    query_properties = {}
    resource_id_format = "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/disks/{disk_name}"
    resource_idem_path = "azure.compute.disks"
    api_version = hub.tool.azure.api_versions.compute.disks
    return await hub.tool.azure.generic.run_absent(
        ctx,
        resource_idem_path,
        api_version,
        name,
        resource_id_format,
        resource_id,
        path_properties,
        query_properties,
    )


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Disks under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.compute.disks
    """
    result = {}
    ret_list = await hub.exec.azure.compute.disks.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe compute disks {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.compute.disks.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""States module for managing Key Vault."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 20}}


async def present(
    hub,
    ctx,
    name: str,
    location: str,
    resource_group_name: str,
    vault_name: str,
    sku: make_dataclass(
        "sku",
        [("name", str, field(default=None)), ("family", str, field(default=None))],
    ) = None,
    tenant_id: str = None,
    soft_delete_retention_days: int = None,
    subscription_id: str = None,
    resource_id: str = None,
    tags: Dict = None,
    enabled_for_deployment: bool = None,
    enabled_for_disk_encryption: bool = None,
    enabled_for_template_deployment: bool = None,
    enable_rbac_authorization: bool = None,
    public_network_access_enabled: str = None,
    purge_protection_enabled: bool = None,
    access_policies: List[
        make_dataclass(
            "accessPolicies",
            [
                ("tenant_id", str, field(default=None)),
                ("object_id", str, field(default=None)),
                ("key_permissions", List[str], field(default=None)),
                ("secret_permissions", List[str], field(default=None)),
                ("storage_permissions", List[str], field(default=None)),
                ("certificate_permissions", List[str], field(default=None)),
            ],
        )
    ] = None,
    network_acls: make_dataclass(
        "networkAcls",
        [
            ("bypass", str, field(default=None)),
            ("default_action", str, field(default=None)),
            ("ip_rules", List[str], field(default=None)),
            ("virtual_network_subnet_ids", List[str], field(default=None)),
        ],
    ) = None,
) -> Dict:
    r"""Create or update key vault.

    Args:

        name(str): The identifier for this state.
        location(str): Resource location. Changing this forces a new resource to be created.
        resource_group_name(str): The name of the resource group.
        vault_name(str): The name of the key vault.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Key vault resource id on Azure
        sku(Dict, Optional): The SKU of the key vault.
        tags(Dict, Optional): Resource tags.
        tenant_id(str, Optional): Tenant id of azure account.
        enabled_for_deployment(bool, Optional): Boolean flag to specify whether Azure Virtual Machines are permitted to retrieve certificates stored as secrets from the key vault.
        enabled_for_disk_encryption(bool, Optional): Boolean flag to specify whether Azure Disk Encryption is permitted to retrieve secrets from the vault and unwrap keys.
        enabled_for_template_deployment(bool, Optional):  Boolean flag to specify whether Azure Resource Manager is permitted to retrieve secrets from the key vault.
        enable_rbac_authorization(bool, Optional): Boolean flag to specify whether Azure Key Vault uses Role Based Access Control (RBAC) for authorization of data actions.
        public_network_access_enabled(str, Optional): Whether public network access is allowed for this Key Vault. Defaults to true
        soft_delete_retention_days(int, Optional): The number of days that items should be retained for once soft-deleted. This value can be between 7 and 90 (the default) days.
        purge_protection_enabled(bool, Optional): Is Purge Protection enabled for this Key Vault?
        access_policies(List[Dict[str, Any]], Optional): Key vault access policies.
        network_acls(Dict[str, Any], Optional): Key vault Network Acl.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_present:
               test-azure-key-vault:
                  azure.key_vault.vault.present:
                      - name: test-key-vault
                      - vault_name: test-key-vault
                      - resource_group_name: azure-resource-group
                      - location: eastus
                      - sku:
                          family: A
                          name: Premium
                      - soft_delete_retention_days: 8
                      - tags:
                          test: test-Ashu
                      - enabled_for_deployment: false
                      - enabled_for_disk_encryption: true
                      - enabled_for_template_deployment: true
                      - enable_rbac_authorization: true
                      - public_network_access_enabled: Enabled
                      - purge_protection_enabled: true
                      - access_policies:
                          - object_id: 00000000-0000-0000-0000-000000000000
                            certificate_permissions:
                              - Get
                              - List
                              - Update
                              - Create
                              - Import
                              - Delete
                              - Recover
                              - Backup
                              - Restore
                              - ManageContacts
                              - ManageIssuers
                              - GetIssuers
                              - ListIssuers
                              - SetIssuers
                              - DeleteIssuers
                            key_permissions:
                              - Get
                              - List
                              - Update
                              - Create
                              - Import
                              - Delete
                              - Recover
                              - Backup
                              - Restore
                              - GetRotationPolicy
                              - SetRotationPolicy
                              - Rotate
                            secret_permissions:
                              - Get
                              - List
                              - Set
                              - Delete
                              - Recover
                              - Backup
                              - Restore
                            tenant_id: 00000000-0000-0000-0000-000000000000
                      - network_acls:
                          bypass: AzureServices
                          default_action: Deny
                          ip_rules: []
                          virtual_network_subnet_ids: []

    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if tenant_id is None:
        tenant_id = ctx.acct.tenant
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.KeyVault/vaults/{vault_name}"
    response_get = await hub.exec.azure.key_vault.vault.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "subscription_id": subscription_id,
                        "location": location,
                        "resource_id": resource_id,
                        "vault_name": vault_name,
                        "sku": sku,
                        "soft_delete_retention_days": soft_delete_retention_days,
                        "purge_protection_enabled": purge_protection_enabled,
                        "tenant_id": tenant_id,
                        "access_policies": access_policies,
                        "network_acls": network_acls,
                        "enabled_for_deployment": enabled_for_deployment,
                        "enabled_for_disk_encryption": enabled_for_disk_encryption,
                        "enabled_for_template_deployment": enabled_for_template_deployment,
                        "enable_rbac_authorization": enable_rbac_authorization,
                        "public_network_access_enabled": public_network_access_enabled,
                        "tags": tags,
                    },
                )
                result["comment"].append(f"Would create azure.key_vault.vault '{name}'")
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.key_vault.vault.convert_present_to_key_vault(
                    location=location,
                    tags=tags,
                    sku=sku,
                    tenant_id=tenant_id,
                    enabled_for_deployment=enabled_for_deployment,
                    enabled_for_disk_encryption=enabled_for_disk_encryption,
                    enabled_for_template_deployment=enabled_for_template_deployment,
                    enable_rbac_authorization=enable_rbac_authorization,
                    public_network_access_enabled=public_network_access_enabled,
                    soft_delete_retention_days=soft_delete_retention_days,
                    purge_protection_enabled=purge_protection_enabled,
                    access_policies=access_policies,
                    network_acls=network_acls,
                )

                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create key vault {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.key_vault.vault.convert_key_vault_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    vault_name=vault_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(f"Created azure.key_vault.vault '{name}'")
                return result

        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.key_vault.vault.convert_key_vault_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                vault_name=vault_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.key_vault.vault.update_key_vault_payload(
                existing_resource, {"tags": tags}
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.key_vault.vault '{name}' doesn't need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.key_vault.vault.convert_key_vault_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        vault_name=vault_name,
                        resource_id=resource_id,
                        subscription_id=subscription_id,
                    )
                    result["comment"].append(
                        f"Would update azure.key_vault.vault '{name}'"
                    )
                return result

            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.key_vault.vault '{name}' doesn't need to be updated."
                )
                return result
            result["comment"] = result["comment"] + new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                success_codes=[200, 201],
                json=new_payload["ret"],
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.key_vault.vault {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.key_vault.vault.convert_key_vault_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                vault_name=vault_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(f"Updated azure.key_vault.vault '{name}'")
            return result
    else:
        hub.log.debug(
            f"Could not get azure.key_vault.vault {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    vault_name: str,
    subscription_id: str = None,
) -> dict:
    r"""Delete a key vault.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        vault_name(str): The name of the key vault.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.key_vault.vault.absent:
                - name: my-kv
                - subscription_id: my-subscription
                - resource_group_name: my-resource-group
                - vault_name: my-fp
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.KeyVault/vaults/{vault_name}"
    response_get = await hub.exec.azure.key_vault.vault.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"]:
            result[
                "old_state"
            ] = hub.tool.azure.key_vault.vault.convert_key_vault_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                vault_name=vault_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.key_vault.vault '{vault_name}'"
                )
                return result
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                success_codes=[200, 202],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.key_vault.vault {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"] = [
                    response_delete["comment"],
                    response_delete["ret"],
                ]
                return result
            result["comment"].append(f"Deleted azure.key_vault.vault '{vault_name}'")
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.key_vault.vault '{vault_name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not azure.key_vault.vault '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all key vault under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.key_vault.vault
    """
    result = {}
    ret_list = await hub.exec.azure.key_vault.vault.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe key_vault vault {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.key_vault.vault.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



"""
idem-azure can autogenerate states as needed using pop-create.
Once autogenerated states are tested, they should be moved from idem_azure/states/azure to idem_azure/states/azure
"""


def __init__(hub):
    # Provides the ctx argument to all execution modules
    # which will have profile info from the account module
    hub.states.azure.ACCT = ["request", "azure"]



import copy
from typing import Any
from typing import Dict


def update_policy_definition_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    if (new_values.get("display_name") is not None) and (
        new_values["display_name"] != existing_properties.get("displayName")
    ):
        new_payload["properties"]["displayName"] = new_values.get("display_name")
        need_update = True
    if (new_values.get("description") is not None) and (
        new_values["description"] != existing_properties.get("description")
    ):
        new_payload["properties"]["description"] = new_values.get("description")
        need_update = True
    if (new_values.get("mode") is not None) and (
        new_values["mode"] != existing_properties.get("mode")
    ):
        new_payload["properties"]["mode"] = new_values.get("mode")
        need_update = True
    if (new_values.get("policy_rule") is not None) and (
        new_values["policy_rule"] != existing_properties.get("policyRule")
    ):
        new_payload["properties"]["policyRule"] = new_values.get("policy_rule")
        need_update = True
    if new_values.get("metadata") is not None:
        excluded_metadata = ["createdOn", "createdBy", "updatedOn", "updatedBy"]
        if existing_properties.get("metadata") is not None:
            existing_properties["metadata"] = {
                k: v
                for k, v in existing_properties["metadata"].items()
                if k not in excluded_metadata
            }
        if new_values["metadata"] != existing_properties.get("metadata"):
            new_payload["properties"]["metadata"] = new_values.get("metadata")
            need_update = True
    # ToDo: to use DeepDiff to compare two Dict once NamespaceDict can be converted to Dict
    if (new_values.get("parameters") is not None) and (
        new_values["parameters"] != existing_properties.get("parameters")
    ):
        new_payload["properties"]["parameters"] = new_values.get("parameters")
        need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_policy_definition_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    policy_definition_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        policy_definition_name: Azure Policy Definition name.
        resource_id: Azure Policy Definition resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "policy_definition_name": policy_definition_name,
        "subscription_id": subscription_id,
    }
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "description": "description",
            "displayName": "display_name",
            "mode": "mode",
            "policyRule": "policy_rule",
            "policyType": "policy_type",
            "parameters": "parameters",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties.get(parameter_raw)
        if properties.get("metadata") is not None:
            excluded_metadata = ["createdOn", "createdBy", "updatedOn", "updatedBy"]
            resource_translated["metadata"] = {
                k: v
                for k, v in properties["metadata"].items()
                if k not in excluded_metadata
            }
    return resource_translated


def convert_present_to_raw_policy_definition(
    hub,
    policy_type: str,
    description: str,
    display_name: str,
    mode: str,
    metadata: Dict = None,
    policy_rule: Dict = None,
    parameters: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        policy_type: The policy type. Possible values are BuiltIn, Custom and NotSpecified.
        description: The description of the policy definition.
        display_name: The display name of the policy definition.
        mode: The policy mode that allows you to specify which resource types will be evaluated. Some examples are All, Indexed, Microsoft.KeyVault.Data.
        metadata: The metadata for the policy definition.
        policy_rule: The policy rule for the policy definition.
        parameters: Parameters for the policy definition.
    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {"displayName": display_name, "policyType": policy_type},
    }
    if description is not None:
        payload["properties"]["description"] = description
    if metadata is not None:
        payload["properties"]["metadata"] = metadata
    if mode is not None:
        payload["properties"]["mode"] = mode
    if policy_rule is not None:
        payload["properties"]["policyRule"] = policy_rule
    if parameters is not None:
        payload["properties"]["parameters"] = {}
        payload["properties"]["parameters"] = parameters
    return payload



def create_comment(hub, resource_type: str, name: str) -> str:
    return f"Created {resource_type} '{name}'"


def would_create_comment(hub, resource_type: str, name: str) -> str:
    return f"Would create {resource_type} '{name}'"


def could_not_create_comment(hub, resource_type: str, name: str) -> str:
    return f"Could not create {resource_type} '{name}'"


def already_exists_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' already exists."


def does_not_exist_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' does not exist."


def no_property_to_be_updated_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' has no property that needs to be updated."


def update_comment(hub, resource_type: str, name: str) -> str:
    return f"Updated {resource_type} '{name}'"


def would_update_comment(hub, resource_type: str, name: str) -> str:
    return f"Would update {resource_type} '{name}'"


def could_not_update_comment(hub, resource_type: str, name: str) -> str:
    return f"Could not update {resource_type} '{name}'"


def could_not_get_comment(hub, resource_type: str, name: str) -> str:
    return f"Could not get {resource_type} '{name}'"


def up_to_date_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' is up to date."


def delete_comment(hub, resource_type: str, name: str) -> str:
    return f"Deleted {resource_type} '{name}'"


def would_delete_comment(hub, resource_type: str, name: str) -> str:
    return f"Would delete {resource_type} '{name}'"


def could_not_delete_comment(hub, resource_type: str, name: str) -> str:
    return f"Could not delete {resource_type} '{name}'"


def already_absent_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' already absent"


def executed_request_comment(
    hub, resource_type: str, name: str, api_method: str
) -> str:
    return f"Exec method '{api_method}' on {resource_type} '{name}'"


def could_not_execute_request_comment(
    hub, resource_type: str, name: str, api_method: str
) -> str:
    return f"Could not exec method '{api_method}' on {resource_type} '{name}'"



"""Tool module for managing Compute Disks."""

RAW_TO_PRESENT_MAP = {
    "subscription_id": "subscriptionId",
    "resource_group_name": "resourceGroupName",
    "disk_name": "diskName",
    "location": "location",
    "creation_data": {
        "create_option": "properties.creationData.createOption",
        "gallery_image_reference": {
            "id": "properties.creationData.galleryImageReference.id",
            "lun": "properties.creationData.galleryImageReference.lun",
        },
        "image_reference": {
            "id": "properties.creationData.imageReference.id",
            "lun": "properties.creationData.imageReference.lun",
        },
        "logical_sector_size": "properties.creationData.logicalSectorSize",
        "security_data_uri": "properties.creationData.securityDataUri",
        "source_resource_id": "properties.creationData.sourceResourceId",
        "source_unique_id": "properties.creationData.sourceUniqueId",
        "source_uri": "properties.creationData.sourceUri",
        "storage_account_id": "properties.creationData.storageAccountId",
        "upload_size_bytes": "properties.creationData.uploadSizeBytes",
    },
    "extended_location": {
        "name": "extendedLocation.name",
        "type": "extendedLocation.type",
    },
    "bursting_enabled": "properties.burstingEnabled",
    "data_access_auth_mode": "properties.dataAccessAuthMode",
    "disk_access_id": "properties.diskAccessId",
    "disk_iops_read_only": "properties.diskIOPSReadOnly",
    "disk_iops_read_write": "properties.diskIOPSReadWrite",
    "disk_mbps_read_only": "properties.diskMBpsReadOnly",
    "disk_mbps_read_write": "properties.diskMBpsReadWrite",
    "disk_size_gb": "properties.diskSizeGB",
    "encryption": {
        "disk_encryption_set_id": "properties.encryption.diskEncryptionSetId",
        "type": "properties.encryption.type",
    },
    "encryption_settings_collection": {
        "enabled": "properties.encryptionSettingsCollection.enabled",
        "encryption_settings": {
            "disk_encryption_key": {
                "secret_url": "properties.encryptionSettingsCollection.encryptionSettings.diskEncryptionKey.secretUrl",
                "source_vault": {
                    "vm_uri": "properties.encryptionSettingsCollection.encryptionSettings.diskEncryptionKey.sourceVault.vmUri"
                },
            },
            "key_encryption_key": {
                "key_url": "properties.encryptionSettingsCollection.encryptionSettings.keyEncryptionKey.keyUrl",
                "source_vault": {
                    "vm_uri": "properties.encryptionSettingsCollection.encryptionSettings.keyEncryptionKey.sourceVault.vmUri"
                },
            },
        },
        "encryption_settings_version": "properties.encryptionSettingsCollection.encryptionSettingsVersion",
    },
    "hyper_v_generation": "properties.hyperVGeneration",
    "max_shares": "properties.maxShares",
    "network_access_policy": "properties.networkAccessPolicies",
    "os_type": "properties.osType",
    "public_network_access": "properties.publicNetworkAccess",
    "security_profile": {
        "secure_vm_disk_encryption_set_id": "properties.securityProfile.secureVMDiskEncryptionSetId",
        "security_type": "properties.securityProfile.securityType",
    },
    "tier": "properties.tier",
    "sku": {"name": "sku.name", "tier": "sku.tier"},
    "tags": "tags",
    "zones": "zones",
}

PRESENT_TO_RAW_MAP = {
    "subscriptionId": "subscription_id",
    "resourceGroupName": "resource_group_name",
    "diskName": "disk_name",
    "location": "location",
    "properties": {
        "creationData": {
            "createOption": "creation_data.create_option",
            "galleryImageReference": {
                "id": "creation_data.gallery_image_reference.id",
                "lun": "creation_data.gallery_image_reference.lun",
            },
            "imageReference": {
                "id": "creation_data.image_reference.id",
                "lun": "creation_data.image_reference.lun",
            },
            "logicalSectorSize": "creation_data.logical_sector_size",
            "securityDataUri": "creation_data.security_data_uri",
            "sourceResourceId": "creation_data.source_resource_id",
            "sourceUniqueId": "creation_data.source_unique_id",
            "sourceUri": "creation_data.source_uri",
            "storageAccountId": "creation_data.storage_account_id",
            "uploadSizeBytes": "creation_data.upload_size_bytes",
        },
        "burstingEnabled": "bursting_enabled",
        "dataAccessAuthMode": "data_access_auth_mode",
        "diskAccessId": "disk_access_id",
        "diskIOPSReadOnly": "disk_iops_read_only",
        "diskIOPSReadWrite": "disk_iops_read_write",
        "diskMBpsReadOnly": "disk_mbps_read_only",
        "diskMBpsReadWrite": "disk_mbps_read_write",
        "diskSizeGB": "disk_size_gb",
        "encryption": {
            "diskEncryptionSetId": "encryption.disk_encryption_set_id",
            "type": "encryption.type",
        },
        "encryptionSettingsCollection": {
            "enabled": "encryptionS_settings_collection.enabled",
            "encryptionSettings": {
                "diskEncryptionKey": {
                    "secretUrl": "encryption_settings_collection.encryption_settings.disk_encryption_key.secret_url",
                    "sourceVault": {
                        "vmUri": "encryption_settings_collection.encryption_settings.disk_encryption_key.source_vault.vm_uri"
                    },
                },
                "keyEncryptionKey": {
                    "keyUrl": "encryption_settings_collection.encryption_settings.key_encryption_key.key_url",
                    "sourceVault": {
                        "vmUri": "encryption_settings_collection.encryption_settings.key_encryption_key.source_vault.vm_uri"
                    },
                },
            },
            "encryptionSettingsVersion": "encryption_settings_collection.encryption_settings_version",
        },
        "hyperVGeneration": "hyper_v_generation",
        "maxShares": "max_shares",
        "networkAccessPolicy": "network_access_policies",
        "osType": "os_type",
        "publicNetworkAccess": "public_network_access",
        "securityProfile": {
            "secureVMDiskEncryptionSetId": "security_profile.secure_vm_disk_encryption_set_id",
            "securityType": "security_profile.security_type",
        },
        "tier": "tier",
    },
    "extendedLocation": {
        "name": "extended_location.name",
        "type": "extended_location.type",
    },
    "sku": {"name": "sku.name", "tier": "sku.tier"},
    "tags": "tags",
    "zones": "zones",
}


def convert_raw_to_present_state(hub, raw_state):
    return hub.tool.azure.generic.convert_state_format(
        raw_state,
        RAW_TO_PRESENT_MAP,
    )


def convert_present_to_raw_state(hub, present_state):
    return hub.tool.azure.generic.convert_state_format(
        present_state,
        PRESENT_TO_RAW_MAP,
    )



import copy
from typing import Any
from typing import Dict


def convert_raw_firewall_policies_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    resource_id: str,
    firewall_policy_name: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        firewall_policy_name: Azure firewall policy resource name.
        resource_id: Azure Virtual Machine resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
      A Dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "firewall_policy_name": firewall_policy_name,
        "location": resource["location"],
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    if "properties" in resource:
        resource_translated["properties"] = resource["properties"]
    if "type" in resource:
        resource_translated["type"] = resource["type"]
    return resource_translated


def convert_present_to_raw_fire_policies(
    hub,
    tags: str,
    location: str,
    dns_settings: Dict[str, Any] = None,
    sku: Dict[str, Any] = None,
    intrusion_detection: Dict[str, Any] = None,
    threat_intelligence_mode: str = None,
    threat_intelligence_allow_list: Dict[str, Any] = None,
    base_policy_id: str = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        tags(Dict, Optional): Resource tags.
        location(str): Resource location. Changing this forces a new resource to be created.
        dns_settings(Dict): Specifies the dns proxy and server of the firewall policy.
        sku(Dict): Specifies the SKU Tier of the Firewall Policy.
        intrusion_detection(Dict): Specifies the intrusion_detection block for Firewall policy.
        threat_intelligence_mode(str):The operation mode for Threat Intelligence. Possible values are Alert, Deny and Off. Defaults to Alert.
        threat_intelligence_allow_list(Dict): Specifies threat_intelligence_allowlist while creating the Firewall policy.
        base_policy_id(str): The ID of the base Firewall Policy.

    Returns:
        A Dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": {}}
    if tags is not None:
        payload["tags"] = tags
    if dns_settings is not None:
        payload["properties"]["dnsSettings"] = convert_present_to_raw_dns(dns_settings)
    if threat_intelligence_allow_list is not None:
        payload["properties"][
            "threatIntelWhitelist"
        ] = convert_present_to_raw_threat_intelligence_allowlist(
            threat_intelligence_allow_list
        )
    if sku is not None:
        payload["properties"]["sku"] = {"tier": sku.get("tier")}
    if intrusion_detection is not None:
        payload["properties"]["intrusionDetection"] = {
            "mode": intrusion_detection.get("mode")
        }
    if base_policy_id is not None:
        payload["properties"]["basePolicy"] = {"id": base_policy_id}
    if threat_intelligence_mode is not None:
        payload["properties"]["threatIntelMode"] = threat_intelligence_mode
    return payload


def convert_present_to_raw_threat_intelligence_allowlist(
    threat_intelligence_allow_list: Dict[str, Any]
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        threat_intelligence_allow_list(list(str)) : List of allowed Threat Intelligence

    Returns:
        Threat white list payload(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    threat_white_list_payload = {}
    if threat_intelligence_allow_list.get("fqdns") is not None:
        threat_white_list_payload["fqdns"] = threat_intelligence_allow_list.get("fqdns")
    if threat_intelligence_allow_list.get("ip_addresses") is not None:
        threat_white_list_payload["ipAddresses"] = threat_intelligence_allow_list.get(
            "ip_addresses"
        )
    return threat_white_list_payload


def convert_present_to_raw_dns(dns_settings: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        dns_settings(list(str)) : List of dns

    Returns:
        Dns list payload(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    dns_payload = {}
    if dns_settings.get("proxy_enabled") is not None:
        dns_payload["enableProxy"] = dns_settings.get("proxy_enabled")
    if dns_settings.get("servers") is not None:
        dns_payload["servers"] = dns_settings.get("servers")
    return dns_payload


def update_fire_policies_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
    PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
        this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
        do an update.

    Returns:
        A result Dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    new_payload.pop("properties", None)
    if need_update:
        result["ret"] = new_payload
    return result



"""Exec module for managing Compute Virtual Machines."""
from collections import OrderedDict
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


RESOURCE_TYPE = "compute.virtual_machines"


async def get(hub, ctx, resource_id: str, raw: bool = False) -> Dict[str, Any]:
    """Get compute virtual machines resource from resource_id.

    Args:
        resource_id(str):
            The resource_id of virtual machine
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.virtual_machines.get resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.compute.virtual_machines.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/virtualMachines/{virtual_machine_name}"

    """
    result = dict(comment=[], result=True, ret=None)
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "virtualMachines": "virtual_machine_name",
        }
    )
    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?$expand=instanceView&api-version={api_version}",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
            if response_get["comment"]:
                result["comment"].append(result["comment"])
        return result

    elif response_get["result"] and response_get["ret"]:
        if raw:
            result["ret"] = response_get["ret"]
        else:
            uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                resource_id, uri_parameters
            )
            result[
                "ret"
            ] = hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
                resource=response_get["ret"],
                idem_resource_name=resource_id,
                resource_id=resource_id,
                **uri_parameter_values,
            )

    return result


async def list_(hub, ctx) -> Dict:
    """List of compute virtual machines

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.virtual_machines.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.compute.virtual_machines.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "virtualMachines": "virtual_machine_name",
        }
    )
    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Compute/virtualMachines?api-version={api_version}",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                result["ret"].append(
                    hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
                        resource=resource,
                        idem_resource_name=resource_id,
                        subscription_id=subscription_id,
                        resource_id=resource_id,
                        **uri_parameter_values,
                    )
                )
    return result


async def power_off(
    hub, ctx, resource_id: str, skip_shutdown: bool = False
) -> Dict[str, Any]:
    """The operation to power off (stop) a virtual machine. The virtual machine can be restarted with the same provisioned resources. You are still charged for this virtual machine.

    Args:
        resource_id(str):
            The resource_id of virtual machine
        skip_shutdown(bool, Optional):
            The parameter to request non-graceful VM shutdown. True value for this flag indicates non-graceful shutdown whereas false indicates otherwise. Default value for this flag is false if not specified

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.virtual_machines.power_off resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.compute.virtual_machines.power_off
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/virtualMachines/{virtual_machine_name}"

    """

    result = await hub.tool.azure.generic.exec_request(
        ctx,
        resource_id=resource_id,
        request_method="post",
        api_method="powerOff",
        query_params={"skipShutdown": skip_shutdown},
        api_version="2022-03-01",
        resource_idem_path="azure.compute.virtual_machines",
    )

    return result


async def start(hub, ctx, resource_id: str) -> Dict[str, Any]:
    """The operation to start a virtual machine.

    Args:
        resource_id(str):
            The resource_id of virtual machine

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.virtual_machines.start resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.compute.virtual_machines.start
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/virtualMachines/{virtual_machine_name}"

    """

    result = await hub.tool.azure.generic.exec_request(
        ctx,
        resource_id=resource_id,
        request_method="post",
        api_method="start",
        api_version="2022-03-01",
        resource_idem_path="azure.compute.virtual_machines",
    )

    return result



import copy
import uuid

from pop.contract import ContractedContext

RESOURCES_WITH_PRESENT_WRAPPER = [
    "azure.sql_database.databases",
    "azure.compute.virtual_machines",
]

LIST_RESOURCES_WITH_ABSENT_WRAPPER = [
    "azure.sql_database.databases",
    "azure.compute.virtual_machines",
]


async def call_present(hub, ctx: ContractedContext):
    r"""Wrapper for present function."""

    name = ctx.kwargs.get("name", None)
    state_ctx = ctx.kwargs.get("ctx") or ctx.args[1]
    assert state_ctx, f"state context is missing: {state_ctx}"

    azure_service_resource_type: str = ctx.ref
    if azure_service_resource_type:
        azure_service_resource_type = azure_service_resource_type.replace("states.", "")

    # TODO: This needs to be removed once all resources follow the contract
    if azure_service_resource_type not in RESOURCES_WITH_PRESENT_WRAPPER:
        return await ctx.func(*ctx.args, **ctx.kwargs)

    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    subscription_id = hub.tool.azure.resource_utils.get_subscription_id_from_account(
        state_ctx, ctx.kwargs.get("subscription_id")
    )
    ctx.kwargs["subscription_id"] = subscription_id

    service_resource_type = azure_service_resource_type.replace("azure.", "")

    hub_ref_exec = hub.exec.azure
    for resource_path_segment in service_resource_type.split("."):
        hub_ref_exec = hub_ref_exec[resource_path_segment]

    resource_id = ctx.kwargs.get("resource_id") or (
        state_ctx.get("rerun_data") or {}
    ).get("resource_id")
    local_params = {**ctx.kwargs}

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    if state_ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.azure.operation_utils.handle_operation(
            state_ctx,
            state_ctx.get("rerun_data"),
            service_resource_type,
            get_resource=False,
        )

        if not handle_operation_ret["result"]:
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
                    result["comment"].extend(handle_operation_ret["comment"])
            else:
                result["result"] = False
                result["comment"].extend(handle_operation_ret["comment"])

            return result

        # operation finished successfully -> get the resource_id to fetch the created/updated resource
        resource_id = state_ctx.get("rerun_data").get("resource_id")
        if not resource_id:
            hub.log.warning(
                f"Resource id not passed in rerun_data for {azure_service_resource_type}"
            )

    if resource_id:
        computed_resource_id = resource_id
    else:
        computed_resource_id = hub.tool.azure.resource_utils.construct_resource_id(
            service_resource_type, local_params
        )
        if not computed_resource_id:
            result["result"] = False
            result["comment"].append(
                f"Could not construct resource ID of {azure_service_resource_type} from input arguments."
            )
            return result

    response_get = await hub_ref_exec.get(state_ctx, resource_id=computed_resource_id)

    if not response_get["result"]:
        result["result"] = False
        result["comment"].append(
            hub.tool.azure.comment_utils.could_not_get_comment(
                azure_service_resource_type, name
            )
        )
        result["comment"].extend(response_get["comment"])
        return result

    if response_get["ret"]:
        # override name as get execs use resource_id as name
        response_get["ret"]["name"] = local_params.get("name")

    # long-running operation has succeeded - both update and create
    if resource_id and state_ctx.get("rerun_data"):
        if not response_get["ret"]:
            result["result"] = False
            result["comment"].append(
                hub.tool.azure.comment_utils.does_not_exist_comment(
                    azure_service_resource_type, name
                )
            )
            result["comment"].extend(response_get["comment"])
            return result

        result["new_state"] = response_get["ret"]
        result["old_state"] = state_ctx.get("rerun_data").get("old_state")

        if result["old_state"]:
            result["comment"].append(
                hub.tool.azure.comment_utils.update_comment(
                    azure_service_resource_type, name
                )
            )
        else:
            result["comment"].append(
                hub.tool.azure.comment_utils.create_comment(
                    azure_service_resource_type, name
                )
            )

        return result

    result["old_state"] = response_get["ret"] or None

    # resource_id passed, so update expected -> resource should exist
    if get_resource_only_with_resource_id and resource_id and (not response_get["ret"]):
        result["result"] = False
        result["comment"].extend(
            [
                hub.tool.azure.comment_utils.does_not_exist_comment(
                    resource_type=f"{azure_service_resource_type}",
                    name=name,
                ),
                *response_get["comment"],
            ]
        )
        return result

    # resource_id not passed, so create expected -> resource should not exist
    if get_resource_only_with_resource_id and (not resource_id) and response_get["ret"]:
        result["result"] = False
        result["comment"].extend(
            [
                hub.tool.azure.comment_utils.already_exists_comment(
                    resource_type=azure_service_resource_type,
                    name=name,
                )
            ]
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    computed_resource_url = hub.tool.azure.resource_utils.construct_resource_url(
        state_ctx,
        service_resource_type,
        resource_id=computed_resource_id,
        subscription_id=subscription_id,
    )

    if not computed_resource_url:
        result["result"] = False
        result["comment"].append(
            f"Could not construct resource URL of {azure_service_resource_type} from input arguments."
        )
        return result

    # TODO: Is this needed?
    if not resource_id and not get_resource_only_with_resource_id:
        resource_id = computed_resource_id

    state_ctx["wrapper_result"] = result
    state_ctx["computed"] = {
        "resource_id": computed_resource_id,
        "resource_url": computed_resource_url,
    }

    return await ctx.func(*ctx.args, **{**ctx.kwargs, "resource_id": resource_id})


async def call_absent(hub, ctx: ContractedContext):
    r"""Wrapper for absent function.

    This method handles the parameters given in the ctx and deletes a resource only if
    the parameters are valid and the resource can be found using them.

    As with the call_present method, here in call_absent we also rely on having rerun_data in order to decide
    whether delete API call needs to be done or we just have to reconcile and wait until for the delete
    operation to complete.

    Steps:
        1. Construct resource_id if it is missing and get_resource_only_with_resource_id flag is NOT set.
            If resource_id is NOT provided and cannot be constructed, then it is directly assumed that the
            resource does not exist in the Cloud. We then return result["result"]=True and result["comment"]
            that the resource is already absent.
        2. Using resource_id check for resource existence by getting the resource current state from the Cloud.
            If the resource is found then delete it or return already absent comment.

    Args:
        hub:
            The redistributed pop central hub. The root of the namespace that pop operates on.
        ctx:
            Invocation context for this command.


    Returns: The result of a resource deletion state.

    """

    state_ctx = ctx.kwargs.get("ctx") or ctx.args[1]
    assert state_ctx, f"state context is missing: {state_ctx}"

    azure_service_resource_type: str = ctx.ref
    if azure_service_resource_type:
        azure_service_resource_type = azure_service_resource_type.replace("states.", "")
        service_resource_type = azure_service_resource_type.replace("azure.", "")

    if azure_service_resource_type not in LIST_RESOURCES_WITH_ABSENT_WRAPPER:
        return await ctx.func(*ctx.args, **ctx.kwargs)

    name = ctx.kwargs.get("name", None)

    result = {
        "comment": [],
        "old_state": state_ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    subscription_id = hub.tool.azure.resource_utils.get_subscription_id_from_account(
        state_ctx, ctx.kwargs.get("subscription_id")
    )
    ctx.kwargs["subscription_id"] = subscription_id

    hub_ref_exec = hub.exec.azure
    for resource_path_segment in service_resource_type.split("."):
        hub_ref_exec = hub_ref_exec[resource_path_segment]

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    resource_id = ctx.kwargs.get("resource_id")
    local_params = {**ctx.kwargs}

    if not resource_id and not get_resource_only_with_resource_id:
        resource_id = (state_ctx.get("old_state") or {}).get(
            "resource_id"
        ) or hub.tool.azure.resource_utils.construct_resource_id(
            service_resource_type, local_params
        )

    if not resource_id and not state_ctx.get("rerun_data"):
        result["comment"].append(
            hub.tool.azure.comment_utils.already_absent_comment(
                azure_service_resource_type, name
            )
        )
        return result

    if not state_ctx.get("rerun_data"):
        get_ret = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not get_ret["result"]:
            result["result"] = False
            result["comment"].append(
                hub.tool.azure.comment_utils.could_not_get_comment(
                    azure_service_resource_type, name
                )
            )
            result["comment"].extend(get_ret["comment"])
            return result

        if not get_ret["ret"]:
            result["result"] = True
            result["comment"].append(
                hub.tool.azure.comment_utils.already_absent_comment(
                    azure_service_resource_type, name
                )
            )
            return result

        result["old_state"] = get_ret["ret"]
        result["old_state"]["name"] = name
    else:
        result["old_state"] = state_ctx["rerun_data"]["old_state"]

    if state_ctx.get("test", False):
        result["comment"].append(
            hub.tool.azure.comment_utils.would_delete_comment(
                azure_service_resource_type, name
            )
        )
        return result

    if not state_ctx.get("rerun_data"):
        resource_url = hub.tool.azure.resource_utils.construct_resource_url(
            state_ctx,
            service_resource_type,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )

        if not resource_url:
            result["result"] = False
            result["comment"].append(
                f"Could not construct resource URL of {azure_service_resource_type} from input arguments."
            )
            return result

        # First iteration; invoke resource's delete()
        response_delete = await hub.exec.request.raw.delete(
            state_ctx,
            url=resource_url,
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            could_not_delete_comment = (
                hub.tool.azure.comment_utils.could_not_delete_comment(
                    azure_service_resource_type, name
                )
            )
            hub.log.debug(could_not_delete_comment)
            result["result"] = False
            result["comment"] = [could_not_delete_comment, response_delete["comment"]]
            return result

        if response_delete["status"] == 202:
            # Deleting the resource is in progress.
            result["rerun_data"] = {
                "operation_id": str(uuid.uuid4()),
                "operation_headers": dict(response_delete.get("headers")),
                "resource_url": resource_url,
                "old_state": result["old_state"],
            }
            return result
        else:
            result["comment"].append(
                hub.tool.azure.comment_utils.delete_comment(
                    azure_service_resource_type, name
                )
            )
            return result
    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.azure.operation_utils.handle_operation(
            state_ctx,
            state_ctx.get("rerun_data"),
            get_resource=True,
        )

        if not handle_operation_ret["result"]:
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
                    result["comment"].extend(handle_operation_ret["comment"])
            else:
                result["result"] = False
                result["comment"].extend(handle_operation_ret["comment"])

            return result

        if handle_operation_ret.get("resource"):
            # Get for resource returned an object
            result["result"] = False
            could_not_delete_comment = (
                hub.tool.azure.comment_utils.could_not_delete_comment(
                    azure_service_resource_type, name
                )
            )
            result["comment"] = [could_not_delete_comment, "Resource might still exist"]
            result["comment"].extend(handle_operation_ret["comment"])
            return result

        result["comment"].append(
            hub.tool.azure.comment_utils.delete_comment(
                azure_service_resource_type, name
            )
        )

        return result



"""State module for managing Network interface."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
RESOURCE_TYPE = "network.network_interfaces"


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_interface_name: str,
    location: str,
    ip_configurations: List[
        make_dataclass(
            "IpConfigurationsSet",
            [
                ("name", str, field(default=None)),
                ("private_ip_address_allocation", str, field(default=None)),
                ("subnet_id", str, field(default=None)),
                ("private_ip_address_version", str, field(default=None)),
                ("private_ip_address", str, field(default=None)),
                ("public_ip_address_id", str, field(default=None)),
                ("primary", bool, field(default=False)),
            ],
        )
    ],
    enable_accelerated_networking: bool = None,
    dns_settings: make_dataclass(
        "NetworkInterfaceDnsSettings",
        [
            ("dns_servers", List[str], field(default=None)),
            ("internal_dns_name_label", str, field(default=None)),
        ],
    ) = None,
    network_security_group_id: str = None,
    tags: Dict = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Network Interfaces.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_interface_name(str): The name of the network interface.
        location(str): Resource location. This field can not be updated.
        ip_configurations(List[Dict[str, Any]]): A list of IPConfigurations of the network interface. Each ip configuration supports fields:

            * name(str):
                The name of the resource that is unique within a resource group.
            * private_ip_address_allocation(str):
                The allocation method used for the Private IP Address. Possible values are Dynamic and Static.
                Azure does not assign a Dynamic IP Address until the Network Interface is attached to a running Virtual Machine(or other resource).
            * subnet_id(str):
                Resource ID of the Subnet bound to the IP configuration. The name of the resource that is unique within a resource group.
                This is required when private_ip_address_version is set to IPv4.
            * private_ip_address_version(str):
                The specific IP configuration is IPv4 or IPv6. Default is IPv4.
            * private_ip_address(str):
                The Static IP Address which should be used. When private_ip_address_allocation is set to Static, private_ip_address can be configured.
            * public_ip_address_id(str):
                Resource ID of the Public IP address bound to the IP configuration. The name of the resource that is unique within a resource group.
            * primary(bool):
                To check if this is the primary IP Configuration. Must be true for the first ip_configuration when
                multiple are specified. Defaults to false. Primary attribute must be true for the first ip_configuration
                when multiple are specified. Defaults to false.
        enable_accelerated_networking(bool, Optional): If the network interface is accelerated networking enabled.
        dns_settings(Dict[str, Any], Optional): The DNS settings in network interface.

            * dns_servers(List[str, Any], Optional):
                List of DNS servers IP addresses. Use 'AzureProvidedDNS' to switch to azure provided DNS resolution. 'AzureProvidedDNS' value cannot be combined with other IPs, it must be the only value in dnsServers collection.
            * internal_dns_name_label(str, Optional):
                Relative DNS name for this NIC used for internal communications between VMs in the same virtual network.

        network_security_group_id(str, Optional): ID of the network security group.
        tags(Dict[str, str], Optional): Resource tags.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Network Interface resource id on Azure.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            my-nic:
              azure.network.network_interfaces.present:
                - name: my-nic
                - resource_group_name: my-rg
                - network_interface_name: my-nic
                - location: southindia
                - subscription_id: my-subscription
                - ip_configurations:
                  - name: my-ipc
                    private_ip_address_allocation: Static
                    subnet_id: subnet_name
                    private_ip_address_version: IPv4
                    private_ip_address: 10.0.0.24
                    primary: true
                - tags:
                    my-tag-key: my-tag-value

    """
    path_properties = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_group_name,
        "network_interface_name": network_interface_name,
    }
    query_properties = {}
    body_properties = {
        "location": location,
        "ip_configurations": ip_configurations,
        "tags": tags,
        "enable_accelerated_networking": enable_accelerated_networking,
        "dns_settings": dns_settings,
        "network_security_group_id": network_security_group_id,
    }
    resource_idem_path = "azure.network.network_interfaces"
    resource_id_format = "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkInterfaces/{network_interface_name}"
    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    result = await hub.tool.azure.generic.run_present(
        ctx,
        resource_idem_path,
        api_version,
        name,
        resource_id,
        resource_id_format,
        path_properties,
        query_properties,
        body_properties,
    )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    network_interface_name: str = None,
    subscription_id: str = None,
) -> Dict:
    r"""Delete Network Interfaces.

    Args:
        name(str): The identifier for this state.
        resource_id(str, Optional): Network Interface resource id in Azure.
        resource_group_name(str): The name of the resource group.
        network_interface_name(str): The name of the network interface.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.network_interfaces.absent:
                - name: value
                - resource_group_name: value
                - network_interface_name: value
                - subscription_id: my-subscription
    """
    path_properties = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_group_name,
        "network_interface_name": network_interface_name,
    }
    query_properties = {}
    resource_idem_path = "azure.network.network_interfaces"
    resource_id_format = "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkInterfaces/{network_interface_name}"
    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    result = await hub.tool.azure.generic.run_absent(
        ctx,
        resource_idem_path,
        api_version,
        name,
        resource_id_format,
        resource_id,
        path_properties,
        query_properties,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Network Interfaces under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.network_interfaces
    """
    result = {}
    ret_list = await hub.exec.azure.network.network_interfaces.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe network_interfaces {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.network_interfaces.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



from typing import List

try:
    import bs4
    import requests
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def parse_file(hub, file_path: str, base_url="https://docs.microsoft.com"):
    """
    Parse a page such as this one: https://docs.microsoft.com/en-us/rest/api/?view=Azure

    Create a pretty version of a list of Azure services"
    """
    hub.log.warning(
        "This function should be deprecated and replaced with parsing the URL directly"
    )
    with open(file_path) as f:
        html = f.read()
    soup = bs4.BeautifulSoup(html, "lxml")

    return soup


def parse_url(hub, url: str):
    """
    Parse a page such as this one: https://docs.microsoft.com/en-us/rest/api/?view=Azure

    Create a pretty version of a list of Azure services"
    """
    with requests.get(
        url, params={"textDecorations": True, "textFormat": "HTML"}
    ) as response:
        # Evaluate js on page
        html = response.text
        if response.status_code != 200:
            return {}

    assert "Advisor" in html
    soup = bs4.BeautifulSoup(html, parser="lxml")

    return soup


def parse(hub, soup: bs4.BeautifulSoup, resources: List[str] = None):
    base_url = "https://docs.microsoft.com"
    plugins = {}

    # Parse page and get links, names, and descriptions of each Azure service
    parsing_results = []
    api_table = soup.find(class_="api-search-results")
    table_body = api_table.find("tbody")
    for row in table_body.find_all("tr"):
        try:
            cols = row.find_all("td")
            name = cols[0].find("a").text
            if resources and name not in resources:
                hub.log.debug(f"Skipping resource: {name}")
                continue

            parsing_result = {
                "name": name,
                "url": base_url + cols[0].find("a")["href"],
                "description": cols[1].text,
            }
            parsing_results.append(parsing_result)
        except [AttributeError, KeyError] as parse_error:
            hub.log.error.error(
                f"Fail to parse table from url {base_url} at row {str(row)}. Error: {parse_error}"
            )

    progress = tqdm.tqdm(parsing_results)
    for api_reference_data in progress:
        name = api_reference_data["name"]
        progress.set_description(name)
        plugins[name] = hub.pop_create.azure.rest.parse(
            url=api_reference_data["url"], desc=api_reference_data["description"]
        )

    return plugins



"""Exec module for managing Network Interfaces"""

__func_alias__ = {"list_": "list"}

from typing import Dict
from collections import OrderedDict


RESOURCE_TYPE = "network.network_interfaces"


async def get(hub, ctx, resource_id: str, raw: bool = False) -> Dict:
    """Gets Network interface from azure account.

    Args:
        resource_id(str):
            The resource id of the Network interface.
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.network_interfaces.get resource_id="value" raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.network_interfaces.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkInterfaces/{network_interface_name}"
                    raw: False
    """
    result = dict(comment=[], ret=None, result=True)

    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"] = response_get["comment"]
        return result
    if raw:
        result["ret"] = response_get["ret"]
    else:
        uri_parameters = OrderedDict(
            {
                "subscriptions": "subscriptionId",
                "resourceGroups": "resourceGroupName",
                "networkInterfaces": "networkInterfaceName",
            }
        )
        uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
            resource_id, uri_parameters
        )
        result[
            "ret"
        ] = hub.tool.azure.network.network_interfaces.convert_raw_to_present_state(
            {**uri_parameter_values, **response_get["ret"]}
        )

        result["ret"]["name"] = resource_id
        result["ret"]["resource_id"] = resource_id
    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Lists all Network Interface.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.network_interfaces.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.network.network_interfaces.list

    """
    result = dict(comment=[], ret=[], result=True)
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscriptionId",
            "resourceGroups": "resourceGroupName",
            "networkInterfaces": "networkInterfaceName",
        }
    )
    api_version = hub.tool.azure.api_versions.get_api_version(RESOURCE_TYPE)
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/networkInterfaces?api-version={api_version}",
        success_codes=[200],
    ):
        resource_list = page_result.get("value", None)
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                present_state = hub.tool.azure.network.network_interfaces.convert_raw_to_present_state(
                    {**uri_parameter_values, **resource}
                )

                present_state["resource_id"] = resource_id
                present_state["name"] = resource_id

                result["ret"].append(present_state)
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_raw_firewall_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    resource_id: str,
    firewall_name: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        firewall_name: Azure firewall resource name.
        resource_id: Azure firewall resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
      A Dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "firewall_name": firewall_name,
        "location": resource["location"],
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    if "properties" in resource:
        resource_translated["properties"] = resource["properties"]
    if "type" in resource:
        resource_translated["type"] = resource["type"]
    if "zones" in resource:
        resource_translated["zones"] = resource["zones"]
    return resource_translated


def convert_present_to_raw_firewall(
    hub,
    tags: str,
    location: str,
    zones: str,
    sku: Dict[str, Any] = None,
    firewall_policy_id: str = None,
    ip_configuration: List[Dict[str, Any]] = None,
    management_ip_configuration: Dict[str, Any] = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        tags(Dict, Optional): Resource tags.
        location(str): Resource location. Changing this forces a new resource to be created.
        zones(str): A list of availability zones denoting where the resource needs to come from.
        sku(Dict, Optional): The SKU of the Firewall.
        firewall_policy_id(str): The ID of the Firewall Policy applied to this Firewall.
        ip_configuration(List(Dict), Optional): IP configuration of the Firewall.
        management_ip_configuration(Dict, Optional): Management IP configuration of the Firewall.
    Returns:
        A Dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location, "properties": {}}
    if tags is not None:
        payload["tags"] = tags
    if zones is not None:
        payload["zones"] = zones
    if sku is not None:
        payload["properties"]["sku"] = {
            "name": sku.get("name"),
            "tier": sku.get("tier"),
        }
    if firewall_policy_id is not None:
        payload["properties"]["firewallPolicy"] = {"id": firewall_policy_id}
    if ip_configuration is not None:
        payload["properties"]["ipConfigurations"] = convert_present_to_ip_configuration(
            ip_configuration
        )
    if management_ip_configuration is not None:
        payload["properties"][
            "managementIpConfiguration"
        ] = convert_present_to_management_ip_configuration(management_ip_configuration)
    return payload


def convert_present_to_management_ip_configuration(
    management_ip_configuration: Dict[str, Any]
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        management_ip_configuration(Dict[str, Any]) : List of ip config payload for VM

    Returns:
        List of Management ip configurations payload Dict[str,any] in the format of an Azure PUT operation payload.
    """

    management_ip_config_payload = {
        "name": management_ip_configuration.get("name"),
        "properties": {
            "subnet": {"id": management_ip_configuration.get("subnet_id")},
            "publicIPAddress": {
                "id": management_ip_configuration.get("public_ip_address_id")
            },
        },
    }
    return management_ip_config_payload


def convert_present_to_ip_configuration(ip_configurations: List[Dict[str, Any]]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        ip_configurations(list(Dict[str, Any]) : List of ip config payload for VM

    Returns:
        List of ip configurations payload List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    data_ip_configuration_list: List = []
    for ip_config in ip_configurations:
        ip_config_payload = {
            "name": ip_config.get("name"),
            "properties": {
                "subnet": {"id": ip_config.get("subnet_id")},
                "publicIPAddress": {"id": ip_config.get("public_ip_address_id")},
            },
        }
        data_ip_configuration_list.append(ip_config_payload)
    return data_ip_configuration_list


def update_firewall_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
    PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
        this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
        do an update.

    Returns:
        A result Dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    new_payload.pop("properties", None)
    if need_update:
        result["ret"] = new_payload
    return result



"""State module for managing storage blob."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    account_name: str,
    container_name: str,
    resource_group_name: str,
    subscription_id: str = None,
    resource_id: str = None,
    default_encryption_scope: str = None,
    deny_encryption_scope_override: bool = False,
    has_immutability_policy: bool = False,
    immutable_storage_with_versioning: bool = False,
    metadata: Dict = None,
    public_access: str = "Blob",
    remaining_retention_days: int = 0,
):
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = (
            f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Storage/"
            f"storageAccounts/{account_name}/blobServices/default/containers/{container_name}"
        )

    response_get = await hub.exec.azure.storage_resource_provider.storage_blob.get(
        ctx, resource_id=resource_id
    )

    if response_get["result"]:
        # 1. Resource is absent,
        if not response_get["ret"]:
            # Test mode
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "account_name": account_name,
                        "container_name": container_name,
                        "subscription_id": subscription_id,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.storage_resource_provider.blob '{name}'"
                )
                return result
            # Create resource
            else:
                # PUT operation to create a resource
                payload = {
                    "id": resource_id,
                    "name": name,
                    "container_name": container_name,
                    "type": "Microsoft.Storage/storageAccounts/blobServices/containers",
                    "properties": {
                        "metadata": metadata,
                        "deleted": False,
                        "remainingRetentionDays": remaining_retention_days,
                        "publicAccess": public_access,
                        "hasImmutabilityPolicy": has_immutability_policy,
                        "hasLegalHold": False,
                    },
                }

                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                    success_codes=[200, 201],
                    json=payload,
                )

                if response_put["status"] == 404:
                    hub.log.debug(
                        f"Could not create azure.storage_resource_provider.storage_blob {response_put['comment']} {response_put['ret']}"
                    )
                    result["result"] = False
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    return result
                elif response_put["status"] == 202:
                    get_response = {"result": False}
                    while not get_response["result"]:
                        get_response = await hub.exec.request.json.get(
                            ctx,
                            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                            success_codes=[200],
                        )
                    result[
                        "new_state"
                    ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                        subscription_id=subscription_id,
                        resource_grp_name=resource_group_name,
                        account_name=account_name,
                        result_dict=get_response["ret"],
                    )

                elif not response_put["result"] and response_put["status"] != 202:
                    hub.log.debug(
                        f"Could not create azure.storage_resource_provider.storage_blob {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                else:
                    result_payload = get_create_payload(
                        resource_id,
                        name,
                        container_name,
                        metadata,
                        resource_group_name,
                        remaining_retention_days,
                        public_access,
                        has_immutability_policy,
                    )
                    result[
                        "new_state"
                    ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                        subscription_id=subscription_id,
                        resource_grp_name=resource_group_name,
                        account_name=account_name,
                        result_dict=result_payload,
                    )

                result["comment"].append(
                    f"Created azure.storage_resource_provider.storage_blob '{name}'"
                )
                return result

        # 2. Resource is present,
        else:
            existing_resource = response_get["ret"]
            # Create old and new payloads
            result[
                "old_state"
            ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                subscription_id=subscription_id,
                resource_grp_name=resource_group_name,
                account_name=account_name,
                result_dict=existing_resource,
            )
            changed_resources = {
                "default_encryption_scope": default_encryption_scope,
                "deny_encryption_scope_override": deny_encryption_scope_override,
                "has_immutability_policy": has_immutability_policy,
                "immutable_storage_with_versioning": immutable_storage_with_versioning,
                "metadata": metadata,
                "remaining_retention_days": remaining_retention_days,
                "public_access": public_access,
            }
            # Generate PUT operation payload with new values
            new_payload = hub.tool.azure.storage_resource_provider.storage_blob.update_storage_blob_payload(
                existing_resource, changed_resources
            )

            # Test mode
            if ctx.get("test", False):
                # i. If no changed property found which needs to be updated.
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.storage_resource_provider.storage_blob '{name}' has no property need to be updated."
                    )
                # ii. Properties will be updated
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                        subscription_id=subscription_id,
                        resource_grp_name=resource_group_name,
                        account_name=account_name,
                        result_dict=new_payload["ret"],
                    )
                    result["comment"].append(
                        f"Would update azure.storage_resource_provider.storage_blob '{name}'"
                    )
                return result

            # PUT operation nothing to update the resource
            if new_payload["ret"] == None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.storage_resource_provider.storage_blob '{name}' has no property need to be updated."
                )
                return result

            # Call PUT operation to update the resource
            result["comment"] = result["comment"] + new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                success_codes=[200],
                json=new_payload["ret"],
            )
            if response_put["result"] == False:
                hub.log.debug(
                    f"Could not update azure.storage_resource_provider.storage_blob {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                subscription_id=subscription_id,
                resource_grp_name=resource_group_name,
                account_name=account_name,
                result_dict=response_put["ret"],
            )
            if result["old_state"] == result["new_state"]:
                result["comment"].append(
                    f"azure.storage_resource_provider.storage_blob '{name}' has no property need to be updated."
                )
                return result

            result["comment"].append(
                f"Updated azure.storage_resource_provider.storage_blob '{name}'"
            )
            return result
    # 3. If issue occurred while finding resource
    else:
        hub.log.debug(
            f"Could not get azure.storage_resource_provider.storage_blob {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    account_name: str,
    container_name: str,
    subscription_id: str = None,
):
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = (
        f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Storage/"
        f"storageAccounts/{account_name}/blobServices/default/containers/{container_name}"
    )
    response_get = await hub.exec.azure.storage_resource_provider.storage_blob.get(
        ctx,
        resource_id=resource_id,
    )
    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.storage_resource_provider.storage_blob '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result

    if response_get["ret"]:
        result["old_state"] = response_get["ret"]
        result["old_state"]["name"] = name

        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.storage_resource_provider.storage_blob '{name}'"
            )
            return result
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 202],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.storage_resource_provider.storage_blob {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(
            f"Deleted azure.storage_resource_provider.storage_blob '{name}'"
        )
        return result
    else:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.storage_resource_provider.storage_blob '{name}' already absent"
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret_list = await hub.exec.azure.storage_resource_provider.storage_blob.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe storage account {ret_list['comment']}")
        return result
    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.storage_resource_provider.storage_blob.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def get_create_payload(
    p_resource_id,
    p_name,
    p_container_name,
    p_metadata,
    resource_group_name,
    p_remaining_retention_days,
    p_public_access,
    p_has_immutability_policy,
):
    return {
        "id": p_resource_id,
        "name": p_name,
        "container_name": p_container_name,
        "resource_group_name": resource_group_name,
        "type": "Microsoft.Storage/storageAccounts/blobServices/containers",
        "properties": {
            "metadata": p_metadata,
            "deleted": False,
            "remainingRetentionDays": p_remaining_retention_days,
            "publicAccess": p_public_access,
            "hasImmutabilityPolicy": p_has_immutability_policy,
            "hasLegalHold": False,
        },
    }



from typing import Any
from typing import Dict
from typing import List


def cleanup_none_values(hub, input_dict: Dict):
    return {k: v for (k, v) in input_dict.items() if v is not None}


def dict_add_nested_key_value_pair(
    hub, input_dict: Dict[str, Any], key_path: List[str], value
):
    recurse_dict = input_dict
    for idx, key in enumerate(key_path):
        if idx == len(key_path) - 1:
            recurse_dict[key] = value
            break
        recurse_dict.setdefault(key, {})
        recurse_dict = recurse_dict[key]
    return input_dict



def parse(hub, ctx, resource_spec: dict, shared_function_data: dict):
    create_api_spec = hub.pop_create.azure.api_spec.parse(resource_spec.get("create"))
    delete_api_spec = hub.pop_create.azure.api_spec.parse(resource_spec.get("delete"))
    list_api_spec = hub.pop_create.azure.api_spec.parse(resource_spec.get("list"))
    patch_parameters = dict()
    if "update" in resource_spec:
        update_api_spec = hub.pop_create.azure.api_spec.parse(
            resource_spec.get("update")
        )
        patch_parameters = generate_patch_parameters(hub, update_api_spec)
    describe_parameters = generate_describe_parameters(hub, create_api_spec)
    present_parameters = generate_parameters(hub, create_api_spec)
    present_parameters["resource_id"] = {
        "default": None,
        "doc": "The identifier for this state on Azure",
        "param_type": "str",
        "required": False,
        "target": "hardcoded",
        "target_type": "arg",
    }
    present = {
        "params": present_parameters,
        "doc": f"Create or update {resource_spec.get('resource')}",
        "hardcoded": dict(
            path=create_api_spec.get("api_url"),
            patch_parameters=patch_parameters,
            **shared_function_data,
        ),
    }
    absent_parameters = generate_parameters(hub, delete_api_spec)
    absent = {
        "params": absent_parameters,
        "doc": f"Delete {resource_spec.get('resource')}",
        "hardcoded": dict(path=delete_api_spec.get("api_url"), **shared_function_data),
    }
    describe = {
        "doc": f"List all {resource_spec.get('resource')} under the same subscription",
        "hardcoded": dict(
            path=list_api_spec.get("api_url"),
            describe_parameters=describe_parameters,
            **shared_function_data,
        ),
    }
    get = {
        "params": {"name": hub.pop_create.azure.template.NAME_PARAMETER},
        "doc": f"Get a single {resource_spec.get('resource')}",
        "hardcoded": dict(path=list_api_spec.get("api_url"), **shared_function_data),
    }
    doc_description = ""
    plugin = {
        "imports": ["from typing import *", "import copy"],
        "contracts": ["resource"],
        "doc": doc_description,
        "functions": {
            "present": present,
            "absent": absent,
            "describe": describe,
            "get": get,
            "list": describe,
            "create": present,
            "update": present,
            "delete": absent,
        },
    }
    if ctx.create_plugin == "auto_states":
        plugin["contracts"] = ["auto_state", "soft_fail"]
    return plugin


def generate_parameters(hub, api_spec: dict):
    params = {}
    api_url = api_spec.get("api_url")
    uri_parameters = api_spec.get("uri_parameters")
    for uri_parameter in uri_parameters:
        params["name"] = hub.pop_create.azure.template.NAME_PARAMETER
        # Skip api-version parameter as this is not an input field. Skip subscriptionId parameter as this will
        # be supplied by acct
        if (
            uri_parameter.get("name") == "api-version"
            or uri_parameter.get("name") == "subscriptionId"
        ):
            continue
        if uri_parameter.get("name_formatted") in api_url:
            param = {
                "required": True,
                "default": None,
                "target_type": "mapping",
                "target": "kwargs",
                "param_type": "Text",
                "doc": uri_parameter.get("description"),
            }
            params.update({uri_parameter.get("name_formatted"): param})
    return params


def generate_patch_parameters(hub, api_spec: dict):
    patch_parameters = {}
    request_body = api_spec.get("request_body")
    for parameter in request_body:
        if "." in parameter.get("name"):
            parameter_split = parameter.get("name").split(".")
            if parameter_split[0] not in patch_parameters:
                patch_parameters[parameter_split[0]] = dict()
            patch_parameters[parameter_split[0]][parameter_split[1]] = parameter_split[
                1
            ]
        else:
            patch_parameters[parameter.get("name")] = parameter.get("name")
    return patch_parameters


def generate_describe_parameters(hub, api_spec: dict):
    params = {}
    api_url = api_spec.get("api_url")
    uri_parameters = api_spec.get("uri_parameters")
    # remove api-version parameter from uri
    uri_values = api_url.split("?")[0].split("/")
    for uri_parameter in uri_parameters:
        # skip api-version and subscriptionId parameter
        if (
            uri_parameter.get("name") == "api-version"
            or uri_parameter.get("name") == "subscriptionId"
        ):
            continue
        if uri_parameter.get("name_formatted") in api_url:
            uri_index = uri_values.index(
                "{" + uri_parameter.get("name_formatted") + "}"
            )
            params.update(
                {uri_values[uri_index - 1]: uri_parameter.get("name_formatted")}
            )
    return params



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

# TODO: add generic logic which takes matching values by default,
#   also between snake and camel case conversions
#   needs override only for special cases - renamings
PRESENT_TO_RAW_MAP = {
    "properties": {
        "licenseType": "license_type",
        "diagnosticsProfile": {
            "bootDiagnostics": {
                "enabled": "boot_diagnostics.enabled",
                "storageUri": "boot_diagnostics.storage_uri",
            }
        },
        "availabilitySet": {"id": "availability_set_id"},
        "osProfile": {
            "adminUsername": "os_profile.admin_username",
            "adminPassword": "os_profile.admin_password",
            "computerName": "os_profile.computer_name",
            "customData": "os_profile.custom_data",
            "linuxConfiguration": {
                "ssh": {
                    "publicKeys": [
                        {
                            "$list_refs": [
                                "os_profile.linux_configuration.ssh_public_keys"
                            ],
                            "keyData": "os_profile.linux_configuration.ssh_public_keys.key_data",
                            "path": "os_profile.linux_configuration.ssh_public_keys.path",
                        }
                    ]
                },
                "provisionVMAgent": "os_profile.linux_configuration.provision_vm_agent",
                "enableVMAgentPlatformUpdates": "os_profile.linux_configuration.enable_vm_agent_platform_updates",
                "disablePasswordAuthentication": "os_profile.linux_configuration.disable_password_authentication",
            },
            "windowsConfiguration": {
                "provisionVMAgent": "os_profile.windows_configuration.provision_vm_agent",
                "enableVMAgentPlatformUpdates": "os_profile.windows_configuration.enable_vm_agent_platform_updates",
                "enableAutomaticUpdates": "os_profile.windows_configuration.enable_automatic_updates",
                "timeZone": "os_profile.windows_configuration.time_zone",
            },
        },
        "extensionsTimeBudget": "extensions_time_budget",
    },
    "plan": {
        "name": "plan.name",
        "product": "plan.product",
        "promotionCode": "plan.promotion_code",
        "publisher": "plan.publisher",
    },
}

# TODO: add generic logic for converting one map to the other
RAW_TO_PRESENT_MAP = {
    "license_type": "properties.licenseType",
    "boot_diagnostics": {
        "enabled": "properties.diagnosticsProfile.bootDiagnostics.enabled",
        "storage_uri": "properties.diagnosticsProfile.bootDiagnostics.storageUri",
    },
    "plan": {
        "name": "plan.name",
        "product": "plan.product",
        "promotion_code": "plan.promotionCode",
        "publisher": "plan.publisher",
    },
    "availability_set_id": "properties.availabilitySet.id",
    "os_profile": {
        "computer_name": "properties.osProfile.computerName",
        "admin_username": "properties.osProfile.adminUsername",
        "admin_password": "properties.osProfile.adminPassword",
        "custom_data": "properties.osProfile.customData",
        "linux_configuration": {
            "ssh_public_keys": [
                {
                    "$list_refs": [
                        "properties.osProfile.linuxConfiguration.ssh.publicKeys"
                    ],
                    "key_data": "properties.osProfile.linuxConfiguration.ssh.publicKeys.keyData",
                    "path": "properties.osProfile.linuxConfiguration.ssh.publicKeys.path",
                }
            ],
            "provision_vm_agent": "properties.osProfile.linuxConfiguration.provisionVMAgent",
            "enable_vm_agent_platform_updates": "properties.osProfile.linuxConfiguration.enableVMAgentPlatformUpdates",
            "disable_password_authentication": "properties.osProfile.linuxConfiguration.disablePasswordAuthentication",
        },
        "windows_configuration": {
            "provision_vm_agent": "properties.osProfile.windowsConfiguration.provisionVMAgent",
            "enable_vm_agent_platform_updates": "properties.osProfile.windowsConfiguration.enableVMAgentPlatformUpdates",
            "enable_automatic_updates": "properties.osProfile.windowsConfiguration.enableAutomaticUpdates",
            "time_zone": "properties.osProfile.windowsConfiguration.timeZone",
        },
    },
    "extensions_time_budget": "properties.extensionsTimeBudget",
}


def convert_present_to_raw_virtual_machine(
    hub,
    location: str = None,
    network_interface_ids: List[str] = None,
    os_profile: Dict[str, Any] = None,
    storage_os_disk: Dict[str, Any] = None,
    virtual_machine_size: str = None,
    storage_image_reference: Dict[str, Any] = None,
    storage_data_disks: List[Dict[str, Any]] = None,
    tags: Dict = None,
    plan: Dict[str, str] = None,
    availability_set_id: str = None,
    license_type: str = None,
    boot_diagnostics: Dict[str, Any] = None,
    extensions_time_budget: str = None,
    **kwargs,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location(str, Optional): Resource location. Changing this forces a new resource to be created.
        network_interface_ids(List[str], Optional): A list of Network Interface IDs which should be associated with the Virtual Machine.
        os_profile(Dict[str, Any], Optional): Specifies the operating system settings used while creating the virtual machine.
        storage_os_disk(Dict[str, Any], Optional): Specifies information about the operating system disk used by the virtual machine.
        virtual_machine_size(str, Optional): Specifies the size of the Virtual Machine.
        storage_image_reference(Dict[str, Any], Optional): Specifies information about the image to use. Eg- platform images, marketplace images.
        storage_data_disks(List[Dict[str, Any]], Optional): List of Data disks attached/added to a VM.
        tags(Dict, Optional): Resource tags.
        plan(Dict[str, str], Optional): Specifies information about the marketplace image used to create the virtual machine.
        availability_set_id(str, Optional): Specifies id of the availability set that the virtual machine should be assigned to.
        license_type(str, Optional): Specifies that the image or disk that is being used was licensed on-premises.
        boot_diagnostics(Dict[str, Any]): Specifies the boot diagnostic settings state.
        extensions_time_budget(str, Optional): Specifies the time alloted for all extensions to start.

    Returns:
        A Dict in the format of an Azure PUT operation payload.
    """

    expected_ignored = {
        "name",
        "resource_id",
        "resource_group_name",
        "virtual_machine_name",
        "subscription_id",
    }
    if kwargs and kwargs.keys() - expected_ignored:
        hub.log.warning(f"Ignored VM properties: {kwargs.keys() - expected_ignored}!")

    generic_present_state = {
        k: v for (k, v) in locals().items() if k in RAW_TO_PRESENT_MAP.keys()
    }

    generic_raw_state = hub.tool.azure.generic.convert_state_format(
        generic_present_state,
        PRESENT_TO_RAW_MAP,
    )

    payload = generic_raw_state
    if location is not None:
        payload["location"] = location

    if virtual_machine_size is not None:
        hub.tool.azure.utils.dict_add_nested_key_value_pair(
            payload, ["properties", "hardwareProfile", "vmSize"], virtual_machine_size
        )

    if network_interface_ids is not None:
        network_interface_ids_payload = {
            "networkInterfaces": convert_present_to_raw_network_interfaces(
                network_interface_ids
            )
        }
        hub.tool.azure.utils.dict_add_nested_key_value_pair(
            payload, ["properties", "networkProfile"], network_interface_ids_payload
        )

    if storage_image_reference is not None:
        storage_image_reference_payload = hub.tool.azure.compute.virtual_machines.convert_present_to_raw_image_reference(
            storage_image_reference
        )
        hub.tool.azure.utils.dict_add_nested_key_value_pair(
            payload,
            ["properties", "storageProfile", "imageReference"],
            storage_image_reference_payload,
        )

    if storage_os_disk is not None:
        os_disk_payload = convert_present_to_raw_os_disk(storage_os_disk)
        hub.tool.azure.utils.dict_add_nested_key_value_pair(
            payload, ["properties", "storageProfile", "osDisk"], os_disk_payload
        )

    if storage_data_disks is not None:
        data_disks_payload = convert_present_to_raw_data_disks(storage_data_disks)
        hub.tool.azure.utils.dict_add_nested_key_value_pair(
            payload, ["properties", "storageProfile", "dataDisks"], data_disks_payload
        )

    if tags is not None:
        payload["tags"] = tags

    return payload


def convert_raw_virtual_machine_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    virtual_machine_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a Dict that match the format of
    present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        virtual_machine_name: Azure Virtual Machine resource name.
        resource_id: Azure Virtual Machine resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
      A Dict that contains the parameters that match the present function's input format.
    """

    generic_present_state = hub.tool.azure.generic.convert_state_format(
        resource,
        RAW_TO_PRESENT_MAP,
    )

    resource_translated = {
        **generic_present_state,
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "virtual_machine_name": virtual_machine_name,
        "location": resource.get("location"),
        "subscription_id": subscription_id,
    }
    properties = resource.get("properties")
    if properties:
        if properties.get("hardwareProfile") is not None:
            resource_translated["virtual_machine_size"] = properties.get(
                "hardwareProfile"
            ).get("vmSize")

        if properties.get("networkProfile") is not None:
            if properties.get("networkProfile").get("networkInterfaces") is not None:
                network_profile_payload = convert_raw_to_present_network_interface(
                    network_interfaces=properties.get("networkProfile").get(
                        "networkInterfaces"
                    )
                )
                resource_translated["network_interface_ids"] = network_profile_payload

        if properties.get("storageProfile") is not None:
            storage_profile_properties = properties.get("storageProfile")
            if storage_profile_properties.get("imageReference") is not None:
                image_reference_payload = convert_raw_to_present_image_reference(
                    image_reference=storage_profile_properties.get("imageReference")
                )
                resource_translated["storage_image_reference"] = image_reference_payload
            if storage_profile_properties.get("osDisk") is not None:
                os_disk_payload = convert_raw_to_present_os_disk(
                    os_disk=storage_profile_properties.get("osDisk")
                )
                resource_translated["storage_os_disk"] = os_disk_payload
            if storage_profile_properties.get("dataDisks") is not None:
                data_disk_payload = convert_raw_to_present_data_disks(
                    data_disks=storage_profile_properties.get("dataDisks")
                )
                resource_translated["storage_data_disks"] = data_disk_payload

        if properties.get("instanceView"):
            statuses = properties["instanceView"].get("statuses", [])
            power_state_status = next(
                (status for status in statuses if "PowerState" in status["code"]), None
            )
            if power_state_status is not None:
                status = power_state_status["code"].replace("PowerState/", "")
                resource_translated["status"] = status
    if "tags" in resource:
        resource_translated["tags"] = resource.get("tags")

    return resource_translated


def update_virtual_machine_payload(
    hub,
    existing_payload: Dict[str, Any],
    old_values: Dict[str, Any],
    new_values: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
    PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: The existing resource state in raw format. This is usually a GET operation response.
        old_values: The existing resource state in present format
        new_values: A dictionary of desired state values in present format. If any property's value is None,
        this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
        do an update.

    Returns:
        A result Dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}

    generic_update_payload = (
        hub.tool.azure.generic.compute_update_payload_for_key_subset(
            old_values, new_values, key_subset=RAW_TO_PRESENT_MAP.keys()
        )
    )
    need_update = bool(generic_update_payload)

    for old_prop_name, old_prop_value in old_values.items():
        if old_prop_name not in generic_update_payload:
            generic_update_payload[old_prop_name] = copy.deepcopy(old_prop_value)

    new_payload = (
        hub.tool.azure.compute.virtual_machines.convert_present_to_raw_virtual_machine(
            **generic_update_payload
        )
    )

    existing_properties = existing_payload.get("properties", {})

    if new_values.get("virtual_machine_size") is not None:
        if (
            (existing_properties.get("hardwareProfile") is not None)
            and (existing_properties.get("hardwareProfile").get("vmSize") is not None)
            and (
                existing_properties.get("hardwareProfile").get("vmSize")
                != new_values.get("virtual_machine_size")
            )
        ):
            new_virtual_machine_size = new_values.get("virtual_machine_size")
            hub.tool.azure.utils.dict_add_nested_key_value_pair(
                new_payload,
                ["properties", "hardwareProfile", "vmSize"],
                new_virtual_machine_size,
            )
            need_update = True

    if new_values.get("network_interface_ids"):
        if (
            (existing_properties.get("networkProfile") is not None)
            and (
                existing_properties.get("networkProfile").get("networkInterfaces")
                is not None
            )
            and (
                compare_network_interface_payload(
                    existing_properties.get("networkProfile").get("networkInterfaces"),
                    new_values.get("network_interface_ids"),
                )
            )
        ):
            network_profile_payload = {
                "networkInterfaces": convert_present_to_raw_network_interfaces(
                    network_interface_ids=new_values.get("network_interface_ids")
                )
            }
            hub.tool.azure.utils.dict_add_nested_key_value_pair(
                new_payload, ["properties", "networkProfile"], network_profile_payload
            )
            need_update = True
    if new_values.get("storage_data_disks") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (existing_properties.get("storageProfile").get("dataDisks") is not None)
            and (
                compare_storage_data_disks_payload(
                    existing_properties.get("storageProfile").get("dataDisks"),
                    new_values.get("storage_data_disks"),
                )
            )
        ):
            new_data_disks = convert_present_to_raw_data_disks(
                data_disks=merge_data_disks_payloads(
                    convert_raw_to_present_data_disks(
                        existing_properties.get("storageProfile").get("dataDisks")
                    ),
                    new_values.get("storage_data_disks"),
                )
            )
            hub.tool.azure.utils.dict_add_nested_key_value_pair(
                new_payload,
                ["properties", "storageProfile", "dataDisks"],
                new_data_disks,
            )
            need_update = True
    if new_values.get("storage_os_disk") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (existing_properties.get("storageProfile").get("osDisk") is not None)
            and (
                compare_storage_os_disk_payload(
                    existing_properties.get("storageProfile").get("osDisk"),
                    new_values.get("storage_os_disk"),
                )
            )
        ):
            new_os_disk = convert_present_to_raw_os_disk(
                os_disk=merge_dictionary_payloads(
                    convert_raw_to_present_os_disk(
                        existing_properties.get("storageProfile").get("osDisk")
                    ),
                    new_values.get("storage_os_disk"),
                )
            )
            hub.tool.azure.utils.dict_add_nested_key_value_pair(
                new_payload, ["properties", "storageProfile", "osDisk"], new_os_disk
            )
            need_update = True
    if new_values.get("storage_image_reference") is not None:
        if (
            (existing_properties.get("storageProfile") is not None)
            and (
                existing_properties.get("storageProfile").get("imageReference")
                is not None
            )
            and (
                compare_storage_image_reference_payload(
                    existing_properties.get("storageProfile").get("imageReference"),
                    new_values.get("storage_image_reference"),
                )
            )
        ):
            new_image_ref = hub.tool.azure.compute.virtual_machines.convert_present_to_raw_image_reference(
                storage_image_reference=merge_dictionary_payloads(
                    convert_raw_to_present_image_reference(
                        existing_properties.get("storageProfile").get("imageReference")
                    ),
                    new_values.get("storage_image_reference"),
                )
            )
            hub.tool.azure.utils.dict_add_nested_key_value_pair(
                new_payload,
                ["properties", "storageProfile", "imageReference"],
                new_image_ref,
            )
            need_update = True

    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values.get("tags")
        need_update = True

    if need_update:
        result["ret"] = new_payload
    return result


def convert_present_to_raw_network_interfaces(network_interface_ids: List[str]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        network_interface_ids(List[str]) : List of Network Interface Ids

    Returns:
        Network Interface Ids List[Dict[str, Any]] in the format of an Azure PUT operation payload.
    """
    network_interface_id_counter = 0
    network_interfaces_list: List = []
    for network_interface_id in network_interface_ids:
        network_interfaces_payload = {"id": network_interface_id}
        primary_network_payload = {}
        if network_interface_id_counter == 0:
            primary_network_payload["primary"] = True
        else:
            primary_network_payload["primary"] = False
        network_interfaces_payload["properties"] = primary_network_payload
        network_interfaces_list.append(network_interfaces_payload)
        network_interface_id_counter = network_interface_id_counter + 1
    return network_interfaces_list


def convert_present_to_raw_image_reference(
    hub, storage_image_reference: Dict[str, Any]
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        storage_image_reference(Dict(str, Any)) : Specifies information about the image to use in VM creation/update

    Returns:
        Storage Image Reference Dict[str,any] in the format of an Azure PUT operation payload.
    """
    storage_image_reference_payload = {
        "publisher": storage_image_reference.get("image_publisher"),
        "offer": storage_image_reference.get("image_offer"),
        "sku": storage_image_reference.get("image_sku"),
        "version": storage_image_reference.get("image_version"),
        "id": storage_image_reference.get("image_id"),
        "sharedGalleryImageId": storage_image_reference.get("shared_gallery_image_id"),
        "communityGalleryImageId": storage_image_reference.get(
            "community_gallery_image_id"
        ),
    }
    return hub.tool.azure.utils.cleanup_none_values(storage_image_reference_payload)


def convert_present_to_raw_os_disk(os_disk: Dict[str, Any]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        os_disk(Dict(str, Any)) : Specifies information about the operating system disk used by the virtual machine.

    Returns:
        OS Disk Payload Dict[str,any] in the format of an Azure PUT operation payload.
    """
    os_disk_payload = {
        "name": os_disk.get("disk_name"),
        "diskSizeGB": os_disk.get("disk_size_in_GB"),
        "caching": os_disk.get("disk_caching"),
        "createOption": os_disk.get("disk_create_option"),
        "deleteOption": os_disk.get("disk_delete_option"),
        "managedDisk": {
            "id": os_disk.get("disk_id"),
            "storageAccountType": os_disk.get("storage_account_type"),
        },
    }
    if os_disk.get("disk_image_vhd_uri") is not None:
        os_disk_payload["image"] = {"uri": os_disk.get("disk_image_vhd_uri")}
    if os_disk.get("disk_os_type") is not None:
        os_disk_payload["osType"] = os_disk.get("disk_os_type")

    if os_disk.get("disk_vhd_uri") is not None:
        os_disk_payload["vhd"] = {"uri": os_disk.get("disk_vhd_uri")}

    if os_disk.get("write_accelerator_enabled") is not None:
        os_disk_payload["writeAcceleratorEnabled"] = os_disk.get(
            "write_accelerator_enabled"
        )

    return os_disk_payload


def convert_present_to_raw_data_disks(data_disks: List[Dict[str, Any]]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        data_disks(List[Dict[str, Any]] : List of Data Disk payload for VM

    Returns:
        List of Data Disk payload List[Dict[str, Any]] in the format of an Azure PUT operation payload.
    """
    data_disks_list: List = []
    for data_disk in data_disks:
        data_disks_payload = {
            "name": data_disk.get("disk_name"),
            "diskSizeGB": data_disk.get("disk_size_in_GB"),
            "lun": data_disk.get("disk_logical_unit_number"),
            "caching": data_disk.get("disk_caching"),
            "createOption": data_disk.get("disk_create_option"),
            "deleteOption": data_disk.get("disk_delete_option"),
            "managedDisk": {
                "id": data_disk.get("disk_id"),
                "storageAccountType": data_disk.get("storage_account_type"),
            },
        }
        data_disks_list.append(data_disks_payload)
    return data_disks_list


def convert_raw_to_present_network_interface(network_interfaces: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        network_interfaces(List[Dict], optional): Resource List of Network Interfaces in a virtual machine resource.

    Returns:
         A Network Interface Id List that contains the parameters that match respective present function's input format.
    """
    present_network_interfaces: List = []
    for network_interface in network_interfaces:
        present_network_interfaces.append(network_interface.get("id"))
    return present_network_interfaces


def convert_raw_to_present_image_reference(image_reference: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        image_reference(Dict, optional): Image Reference payload in a virtual machine resource.

    Returns:
         Image Reference payload contains the parameters that match respective present function's input format.
    """
    image_reference_payload = {
        "image_sku": image_reference.get("sku"),
        "image_publisher": image_reference.get("publisher"),
        "image_version": image_reference.get("version"),
        "image_offer": image_reference.get("offer"),
        "image_id": image_reference.get("id"),
        "shared_gallery_image_id": image_reference.get("sharedGalleryImageId"),
        "community_gallery_image_id": image_reference.get("communityGalleryImageId"),
    }
    return image_reference_payload


def convert_raw_to_present_os_disk(os_disk: Dict[str, Any]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        os_disk(Dict, optional): OS Disk payload in a virtual machine resource.

    Returns:
         OS Disk payload that contains the parameters that match respective present function's input format.
    """
    os_disk_payload = {
        "disk_name": os_disk.get("name"),
        "disk_caching": os_disk.get("caching"),
    }
    if os_disk.get("diskSizeGB") is not None:
        os_disk_payload["disk_size_in_GB"] = os_disk.get("diskSizeGB")
    if (
        os_disk.get("managedDisk") is not None
        and os_disk.get("managedDisk").get("storageAccountType") is not None
    ):
        os_disk_payload["storage_account_type"] = os_disk.get("managedDisk").get(
            "storageAccountType"
        )
    if (
        os_disk.get("managedDisk") is not None
        and os_disk.get("managedDisk").get("id") is not None
    ):
        os_disk_payload["disk_id"] = os_disk.get("managedDisk").get("id")
    if os_disk.get("createOption") is not None:
        os_disk_payload["disk_create_option"] = os_disk.get("createOption")
    if os_disk.get("deleteOption") is not None:
        os_disk_payload["disk_delete_option"] = os_disk.get("deleteOption")
    if os_disk.get("image", {}).get("uri") is not None:
        os_disk_payload["disk_image_vhd_uri"] = os_disk.get("image").get("uri")
    if os_disk.get("osType") is not None:
        os_disk_payload["disk_os_type"] = os_disk.get("osType")
    if os_disk.get("vhd", {}).get("uri") is not None:
        os_disk_payload["disk_vhd_uri"] = os_disk.get("vhd").get("uri")
    if os_disk.get("writeAcceleratorEnabled") is not None:
        os_disk_payload["write_accelerator_enabled"] = os_disk.get(
            "writeAcceleratorEnabled"
        )
    return os_disk_payload


def convert_raw_to_present_data_disks(data_disks: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a Dict that match the format of
     present input parameters.

    Args:
        data_disks(List[Dict], optional): Resource List of Data Disks in a virtual machine resource.

    Returns:
         A Data Disk List that contains the parameters that match respective present function's input format.
    """
    present_data_disks_payload: List = []
    for data_disk in data_disks:
        data_disk_payload = {
            "disk_name": data_disk.get("name"),
            "disk_logical_unit_number": data_disk.get("lun"),
            "disk_caching": data_disk.get("caching"),
        }
        if data_disk.get("diskSizeGB") is not None:
            data_disk_payload["disk_size_in_GB"] = data_disk.get("diskSizeGB")
        if (
            data_disk.get("managedDisk") is not None
            and data_disk.get("managedDisk").get("storageAccountType") is not None
        ):
            data_disk_payload["storage_account_type"] = data_disk.get(
                "managedDisk"
            ).get("storageAccountType")
        if (
            data_disk.get("managedDisk") is not None
            and data_disk.get("managedDisk").get("id") is not None
        ):
            data_disk_payload["disk_id"] = data_disk.get("managedDisk").get("id")
        if data_disk.get("createOption") is not None:
            data_disk_payload["disk_create_option"] = data_disk.get("createOption")
        if data_disk.get("deleteOption") is not None:
            data_disk_payload["disk_delete_option"] = data_disk.get("deleteOption")
        present_data_disks_payload.append(data_disk_payload)
    return present_data_disks_payload


def compare_network_interface_payload(
    existing_network_interface_list: List[Dict[str, Any]],
    network_interface_ids: List[str],
):
    """
    Compares network interface payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_network_interface_list(List[Dict]): Existing Network Interface payload
        network_interface_ids(List[str]): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    if len(network_interface_ids) != len(existing_network_interface_list):
        return True
    existing_network_interface_id_list = []
    for existing_network_interface in existing_network_interface_list:
        existing_network_interface_id_list.append(existing_network_interface.get("id"))
    return set(existing_network_interface_id_list) != set(network_interface_ids)


def compare_storage_os_disk_payload(
    existing_storage_os_disk_payload: Dict[str, Any],
    new_storage_os_disk: Dict[str, Any],
):
    """
    Compares OS Disk payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_os_disk_payload[Dict]: Existing Storage OS Disk payload
        new_storage_os_disk[Dict]: Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    os_disk_present_converted_payload = convert_raw_to_present_os_disk(
        existing_storage_os_disk_payload
    )
    return compare_update_dict_payload(
        os_disk_present_converted_payload, new_storage_os_disk
    )


def compare_storage_data_disks_payload(
    existing_storage_data_disks_payload: List[Dict[str, Any]],
    new_storage_data_disks: List[Dict[str, Any]],
):
    """
    Compares Data Disks payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_data_disks_payload(List[Dict]): Existing Storage Data Disk payload
        new_storage_data_disks(List[str]): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    data_disks_present_converted_payload = convert_raw_to_present_data_disks(
        existing_storage_data_disks_payload
    )
    if len(data_disks_present_converted_payload) != len(new_storage_data_disks):
        return True
    existing_payload_map = {}
    for element in data_disks_present_converted_payload:
        existing_payload_map[element.get("disk_name")] = element

    return compare_update_dict_list_payload(
        existing_payload_map, new_storage_data_disks
    )


def compare_storage_image_reference_payload(
    existing_storage_image_reference_payload: Dict[str, Any],
    new_storage_image_reference: Dict[str, Any],
):
    """
    Compares Storage Image Reference payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_storage_image_reference_payload(Dict): Existing Storage Image Reference payload
        new_storage_image_reference(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    storage_image_reference_present_converted_payload = (
        convert_raw_to_present_image_reference(existing_storage_image_reference_payload)
    )
    return compare_update_dict_payload(
        storage_image_reference_present_converted_payload, new_storage_image_reference
    )


def compare_update_dict_payload(
    existing_payload: Dict[str, Any], update_payload: Dict[str, Any]
):
    """
    Compares the payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_payload(Dict): Existing payload
        update_payload(Dict): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    for parameter in update_payload:
        if parameter in existing_payload:
            if update_payload.get(parameter) != existing_payload.get(parameter):
                return True
        else:
            return True
    return False


def compare_update_dict_list_payload(
    existing_payload_map: Dict[str, Any], update_payload: List[Dict[str, Any]]
):
    """
    Compares the payload to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        existing_payload_map(Dict): Existing payload Map
        update_payload(List[Dict]): Present value which will be given as input

    Returns:
        A boolean value, True if there is any difference between the arguments else returns False
    """
    for update_element in update_payload:
        disk_name = update_element.get("disk_name")
        if disk_name in existing_payload_map:
            if compare_update_dict_payload(
                existing_payload_map.get(disk_name), update_element
            ):
                return True
        else:
            return True
    return False


def merge_dictionary_payloads(
    existing_payload: Dict[str, Any], update_payload: Dict[str, Any]
):
    """
    Merge the input payload with the payload from Azure Get Response
    Returns the merged payload

    Args:
        existing_payload(Dict): Existing payload From Azure Get API
        update_payload(Dict): Present value which will be given as input

    Returns:
        A Dict of the merged Payload
    """
    existing_payload_copy = copy.deepcopy(existing_payload)
    for update_element_key in update_payload:
        if update_payload.get(update_element_key) is not None:
            existing_payload_copy[update_element_key] = update_payload.get(
                update_element_key
            )
    return existing_payload_copy


def merge_data_disks_payloads(
    existing_payload: List[Dict[str, Any]], update_payload: List[Dict[str, Any]]
):
    """
    Merge the input payload with the payload from Azure Get Response for Data Disks
    Returns the merged Data Disk payload

    Args:
        existing_payload(Dict): Existing payload Map
        update_payload(Dict): Present value which will be given as input

    Returns:
        A Dict of the merged Data Disks Payload
    """
    existing_payload_map = {}
    existing_merged_payload = []
    for element in existing_payload:
        existing_payload_map[element.get("disk_name")] = element
    for update_element in update_payload:
        disk_name = update_element.get("disk_name")
        if disk_name in existing_payload_map:
            # Merge Data
            existing_merged_payload.append(
                merge_dictionary_payloads(
                    existing_payload_map.get(disk_name), update_element
                )
            )
        else:
            # Keep new input data
            existing_merged_payload.append(update_element)
    return existing_merged_payload


def convert_raw_to_present_state(hub, raw_state):
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "virtualMachines": "virtual_machine_name",
        }
    )
    resource_id = raw_state["id"]
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )

    return (
        hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
            resource=raw_state,
            resource_id=resource_id,
            idem_resource_name=raw_state["name"]
            if "name" in raw_state
            else resource_id,
            **uri_parameter_values,
        )
    )


def convert_present_to_raw_state(hub, present_state):
    return (
        hub.tool.azure.compute.virtual_machines.convert_present_to_raw_virtual_machine(
            location=present_state["location"],
            network_interface_ids=present_state["network_interface_ids"],
            os_profile=present_state["os_profile"],
            storage_os_disk=present_state["storage_os_disk"],
            virtual_machine_size=present_state["virtual_machine_size"],
            storage_image_reference=present_state["storage_image_reference"],
            storage_data_disks=present_state["storage_data_disks"],
            tags=present_state["tags"],
            plan=present_state["plan"],
            availability_set_id=present_state["availability_set_id"],
            license_type=present_state["license_type"],
        )
    )



from typing import Any
from typing import Dict


RESOURCE_ID_TEMPLATES = {
    "sql_database.databases": "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Sql/servers/{server_name}/databases/{database_name}",
    "compute.virtual_machines": "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/virtualMachines/{virtual_machine_name}",
}

# TODO: See if all resources share the same URL format
RESOURCE_URL_TEMPLATES = {
    "sql_database.databases": "{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}",
    "compute.virtual_machines": "{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}",
}


def construct_resource_id(hub, resource_type: str, input_props: Dict[str, Any]) -> str:
    input_props = hub.tool.azure.utils.cleanup_none_values(input_props)
    id_template = RESOURCE_ID_TEMPLATES.get(resource_type)
    if not id_template:
        raise ValueError(f"Could not construct resource_id for {resource_type}")

    try:
        return id_template.format(**input_props)
    except:
        return None


def construct_resource_url(
    hub,
    ctx,
    resource_type: str,
    input_props: Dict[str, Any] = None,
    resource_id: str = None,
    subscription_id: str = None,
) -> str:
    url_template = RESOURCE_URL_TEMPLATES.get(resource_type)
    if not url_template:
        raise ValueError(f"Could not construct resource_url for {resource_type}")
    if not input_props:
        input_props = {}
    else:
        input_props = hub.tool.azure.utils.cleanup_none_values(input_props)

    if not resource_id:
        resource_id = input_props.get(
            "resource_id"
        ) or hub.tool.azure.resource_utils.construct_resource_id(
            resource_type, input_props
        )
    if not subscription_id:
        subscription_id = input_props.get("subscription_id") or ctx.acct.subscription_id
    api_version = hub.tool.azure.api_versions.get_api_version(resource_type)

    try:
        return url_template.format(
            **{
                "ctx": ctx,
                **input_props,
                "resource_id": resource_id,
                "subscription_id": subscription_id,
                "api_version": api_version,
            }
        )
    except:
        return None


def get_subscription_id_from_account(
    hub, ctx: Dict, subscription_id: str = None
) -> str:
    """If subscription_id is explicitly passed by the user, this subscription_id will be returned.
    If subscription_id is empty, this method will return default subscription_id from Azure account
    :param hub: Hub
    :param ctx: Context for the execution of the Idem run located in `hub.idem.RUNS[ctx['run_name']]`.
    :param subscription_id: A string explicitly passed by the user.
    :return: The correct subscription_id
    """
    if not subscription_id:
        subscription_id = ctx.get("acct", {}).get("subscription_id")
    if not subscription_id:
        hub.log.warning("Could not find subscription_id in account")
    return subscription_id


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """
    This method enables state specific implementation of is_pending logic,
    based on resource specific attribute(s).
    Usage 'idem state <sls-file> --reconciler=basic', where the reconciler attribute
    can be missed.

    :param hub: The Hub into which the resolved callable will get placed.
    :param ret: The returned dictionary of the last run.
    :param state: The name of the state.
    :param pending_kwargs: (dict, Optional) May include 'ctx' and 'reruns_wo_change_count'.

    :return: True | False
    """
    if not ret:
        return False

    if ret.get("rerun_data") and ret["rerun_data"].get("has_error", False):
        return False

    if ret.get("rerun_data"):
        return True

    if ret["result"]:
        return False

    return (
        pending_kwargs
        and pending_kwargs.get("reruns_wo_change_count", 0)
        <= hub.reconcile.pending.default.MAX_RERUNS_WO_CHANGE
    )
